'''
The LXD modules manage LXD containers
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.lxd import LxdContainers


def get_container_named(name, containers):
    for container in containers:
        if container['name'] == name:
            return container
    else:
        return None


@operation
def container(
    id,
    present=True, image='ubuntu:16.04',
    state=None, host=None,
):
    '''
    Add/remove LXD containers.

    Note: does not check if an existing container is based on the specified
    image.

    + id: name/identifier for the container
    + image: image to base the container on
    + present: whether the container should be present or absent

    Example:

    .. code:: python

        lxd.container(
            name='Add an ubuntu container',
            id='ubuntu19',
            image='ubuntu:19.10',
        )
    '''

    container = get_container_named(id, host.get_fact(LxdContainers))

    # Container exists and we don't want it
    if not present:
        if container:
            if container['status'] == 'Running':
                yield 'lxc stop {0}'.format(id)

            # Command to remove the container:
            yield 'lxc delete {0}'.format(id)
        else:
            host.noop('container {0} does not exist'.format(id))

    # Container doesn't exist and we want it
    if present:
        if not container:
            # Command to create the container:
            yield 'lxc launch {image} {id}'.format(id=id, image=image)
        else:
            host.noop('container {0} exists'.format(id))



from __future__ import unicode_literals

import re
from datetime import datetime


WIN_LS_REGEX = re.compile((
    # filetype and mode
    r'^([darhsl\-]{6})\s+'
    # Windows date
    r'([0-9]{1,2}\/[0-9]{1,2}\/[0-9]{4})\s+([0-9]{1,2}:[0-9]{1,2}\s[AP][M])\s+'
    # Size (Note: no size on directories)
    r'([0-9]+)?\s+'
    # Size and Filename
    r'([\w\/\.@-]+)'
))

WIN_FLAG_TO_TYPE = {
    'd': 'directory',
    '-': 'file',
}

WIN_FLAG_TO_ATTR = {
    'a': 'archive',
    'r': 'readonly',
    'h': 'hidden',
    's': 'system',
    'l': 'link',
    '-': 'none',
}


def _parse_time(time):
    # Try matching windows format
    try:
        tmp = datetime.strptime(time, '%m/%d/%Y %H:%M %p')
        return tmp
    except ValueError:
        pass


def parse_win_ls_output(output, wanted_type):
    if output:
        matches = re.match(WIN_LS_REGEX, output)
        if matches:
            if output[5] == 'l':
                type = 'link'
            else:
                type = WIN_FLAG_TO_TYPE[output[0]]

            if type != wanted_type:
                return False

            hidden = False
            system = False
            readonly = False
            archive = False
            link = False

            mode_without_first = matches.group(1)[1:]

            for c in mode_without_first:
                tmp = WIN_FLAG_TO_ATTR[c]
                if tmp != 'none':
                    if tmp == 'hidden':
                        hidden = True
                    if tmp == 'readonly':
                        readonly = True
                    if tmp == 'archive':
                        archive = True
                    if tmp == 'system':
                        system = True
                    if tmp == 'link':
                        link = True

            mode = {
                'archive': archive,
                'hidden': hidden,
                'readonly': readonly,
                'system': system,
                'link': link,
            }

            date_and_time = '{} {}'.format(matches.group(2), matches.group(3))

            size = '0'
            if type == 'file':
                size = matches.group(4)

            out = {
                'type': type,
                'mode': mode,
                'mtime': _parse_time(date_and_time),
                'size': size,
                'name': matches.group(5),
                # TODO: You will need to run another powershell command to
                #       get the link target, so bailing on that for now.
            }

            return out



from .sysvinit import InitdStatus


class RcdStatus(InitdStatus):
    '''
    Same as ``initd_status`` but for BSD (/etc/rc.d) systems. Unlike Linux/init.d,
    BSD init scripts are well behaved and as such their output can be trusted.
    '''

    command = '''
        for SERVICE in `find /etc/rc.d /usr/local/etc/rc.d -type f`; do
            $SERVICE status 2> /dev/null || $SERVICE check 2> /dev/null
            echo "`basename $SERVICE`=$?"
        done
    '''

    default = dict



# Import `pyinfra_cli` to trigger gevent monkey patching
import pyinfra_cli  # noqa: F401

import logging  # noqa: I100

import gevent.hub


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger('pyinfra.tests')

# Don't print out exceptions inside greenlets (because here we expect them!)
gevent.hub.Hub.NOT_ERROR = (Exception,)



from unittest import TestCase

from mock import MagicMock, patch

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all

from ..util import make_inventory


class TestWinrmConnector(TestCase):
    def setUp(self):
        self.fake_connect_patch = patch('pyinfra.api.connectors.winrm.connect')
        self.fake_connect_mock = self.fake_connect_patch.start()

    def tearDown(self):
        self.fake_connect_patch.stop()

    def test_connect_host(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect(reason=True)
        assert len(state.active_hosts) == 0

    def test_connect_all_password(self):
        inventory = make_inventory(hosts=(
            ('@winrm/somehost', {'winrm_username': 'testuser', 'winrm_password': 'testpass'}),
            ('@winrm/anotherhost', {'winrm_username': 'testuser2', 'winrm_password': 'testpass2'}),
        ))
        state = State(inventory, Config())
        connect_all(state)

        # Get a host
        somehost = inventory.get_host('@winrm/somehost')
        assert somehost.data.winrm_username == 'testuser'
        assert somehost.data.winrm_password == 'testpass'

        assert len(state.active_hosts) == 2

    def test_run_shell_command(self):
        fake_winrm_session = MagicMock()

        fake_winrm = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_winrm.run_cmd.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_winrm_session.return_value = fake_winrm

        inventory = make_inventory(hosts=('@winrm/somehost',))
        State(inventory, Config())
        host = inventory.get_host('@winrm/somehost')
        host.connect()

        command = 'echo hi'

        out = host.run_shell_command(command, stdin='hello', print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        # TODO: assert status is True

        combined_out = host.run_shell_command(
            command, print_output=True,
            return_combined_output=True,
        )
        assert len(combined_out) == 2

        # TODO: fake_winrm.run_shell_command.assert_called_with("'echo hi'", get_pty=False)



from pyinfra.api import FactBase

from .util.packaging import parse_packages


class XbpsPackages(FactBase):
    '''
    Returns a dict of installed XBPS packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'xbps-query -l'
    requires_command = 'xbps-query'

    default = dict

    regex = r'^.. ([a-zA-Z0-9_\-\+]+)\-([0-9a-z_\.]+)'

    def process(self, output):
        return parse_packages(self.regex, output)



from __future__ import unicode_literals

import re

from pyinfra.api import FactBase

from .util.packaging import parse_packages

rpm_regex = r'^(\S+)\ (\S+)$'
rpm_query_format = '%{NAME} %{VERSION}-%{RELEASE}\\n'


class RpmPackages(FactBase):
    '''
    Returns a dict of installed rpm packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'rpm --queryformat "{0}" -qa'.format(rpm_query_format)
    requires_command = 'rpm'

    default = dict

    def process(self, output):
        return parse_packages(
            rpm_regex, output,
            # yum packages are case-sensitive
            lower=False,
        )


class RpmPackage(FactBase):
    '''
    Returns information on a .rpm file:

    .. code:: python

        {
            'name': 'my_package',
            'version': '1.0.0',
        }
    '''

    requires_command = 'rpm'

    def command(self, name):
        return (
            '! test -e {1} || '
            '(rpm --queryformat "{0}" -qp {1} 2> /dev/null || rpm --queryformat "{0}" -q {1})'
        ).format(rpm_query_format, name)

    def process(self, output):
        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                return {
                    'name': matches.group(1),
                    'version': matches.group(2),
                }


class RpmPackageProvides(FactBase):
    '''
    Returns a list of packages that provide the specified capability (command, file, etc).
    '''

    default = list

    requires_command = 'repoquery'

    @staticmethod
    def command(name):
        # Accept failure here (|| true) for invalid/unknown packages
        return 'repoquery --queryformat "{0}" --whatprovides {1} || true'.format(
            rpm_query_format,
            name,
        )

    @staticmethod
    def process(output):
        packages = []

        for line in output:
            matches = re.match(rpm_regex, line)
            if matches:
                packages.append(list(matches.groups()))

        return packages



from __future__ import unicode_literals

import os

from distutils.spawn import find_executable
from tempfile import mkstemp

import click
import six

from pyinfra import logger
from pyinfra.api.exceptions import InventoryError
from pyinfra.api.util import get_file_io

from .util import (
    get_sudo_password,
    make_unix_command,
    run_local_process,
    split_combined_output,
)

EXECUTION_CONNECTOR = True


def make_names_data(hostname=None):
    if hostname is not None:
        raise InventoryError('Cannot have more than one @local')

    yield '@local', {}, ['@local']


def connect(state, host):
    return True


def run_shell_command(
    state, host, command,
    get_pty=False,  # ignored
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    use_sudo_password=False,
    **command_kwargs
):
    '''
    Execute a command on the local machine.

    Args:
        state (``pyinfra.api.State`` object): state object for this command
        host (``pyinfra.api.Host`` object): the target host
        command (string): actual command to execute
        sudo (boolean): whether to wrap the command with sudo
        sudo_user (string): user to sudo to
        env (dict): environment variables to set
        timeout (int): timeout for this command to complete before erroring

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    '''

    if use_sudo_password:
        command_kwargs['use_sudo_password'] = get_sudo_password(
            state, host, use_sudo_password,
            run_shell_command=run_shell_command,
            put_file=put_file,
        )

    command = make_unix_command(command, state=state, **command_kwargs)
    actual_command = command.get_raw_value()

    logger.debug('--> Running command on localhost: {0}'.format(command))

    if print_input:
        click.echo('{0}>>> {1}'.format(host.print_prefix, command), err=True)

    return_code, combined_output = run_local_process(
        actual_command,
        stdin=stdin,
        timeout=timeout,
        print_output=print_output,
        print_prefix=host.print_prefix,
    )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    if return_combined_output:
        return status, combined_output

    stdout, stderr = split_combined_output(combined_output)
    return status, stdout, stderr


def put_file(
    state, host, filename_or_io, remote_filename,
    print_output=False, print_input=False,
    **command_kwargs
):
    '''
    Upload a local file or IO object by copying it to a temporary directory
    and then writing it to the upload location.
    '''

    _, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, 'wb') as temp_f:
                data = file_io.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                temp_f.write(data)

        # Copy the file using `cp` such that we support sudo/su
        status, _, stderr = run_shell_command(
            state, host, 'cp {0} {1}'.format(temp_filename, remote_filename),
            print_output=print_output,
            print_input=print_input,
            **command_kwargs
        )

        if not status:
            raise IOError('\n'.join(stderr))
    finally:
        os.remove(temp_filename)

    if print_output:
        click.echo(
            '{0}file copied: {1}'.format(host.print_prefix, remote_filename),
            err=True,
        )

    return status


def get_file(
    state, host, remote_filename, filename_or_io,
    print_output=False, print_input=False,
    **command_kwargs
):
    '''
    Download a local file by copying it to a temporary location and then writing
    it to our filename or IO object.
    '''

    _, temp_filename = mkstemp()

    try:
        # Copy the file using `cp` such that we support sudo/su
        status, _, stderr = run_shell_command(
            state, host, 'cp {0} {1}'.format(remote_filename, temp_filename),
            print_output=print_output,
            print_input=print_input,
            **command_kwargs
        )

        if not status:
            raise IOError('\n'.join(stderr))

        # Load our file or IO object and write it to the temporary file
        with open(temp_filename) as temp_f:
            with get_file_io(filename_or_io, 'wb') as file_io:
                data = temp_f.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                file_io.write(data)
    finally:
        os.remove(temp_filename)

    if print_output:
        click.echo(
            '{0}file copied: {1}'.format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def check_can_rsync(host):
    if not find_executable('rsync'):
        raise NotImplementedError('The `rsync` binary is not available on this system.')


def rsync(
    state, host, src, dest, flags,
    print_output=False, print_input=False,
    **command_kwargs
):
    status, _, stderr = run_shell_command(
        state, host,
        'rsync {0} {1} {2}'.format(' '.join(flags), src, dest),
        print_output=print_output,
        print_input=print_input,
        **command_kwargs
    )

    if not status:
        raise IOError('\n'.join(stderr))

    return True



'''
Manage pip (python) packages. Compatible globally or inside
a virtualenv (virtual environment).
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.files import File
from pyinfra.facts.pip import PipPackages

from . import files
from .util.packaging import ensure_packages


@operation
def virtualenv(
    path,
    python=None, venv=False, site_packages=False, always_copy=False, present=True,
    state=None, host=None,
):
    '''
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + venv: use standard venv module instead of virtualenv
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    Example:

    .. code:: python

        pip.virtualenv(
            name='Create a virtualenv',
            path='/usr/local/bin/venv',
        )
    '''

    # Check for *contents* of a virtualenv, ie don't accept an empty directory
    # as a valid virtualenv but ensure the activate script exists.
    activate_script_path = '{0}/bin/activate'.format(path)

    if present is False:
        if host.get_fact(File, path=activate_script_path):
            yield files.directory(path, present=False, state=state, host=host)
        else:
            host.noop('virtualenv {0} does not exist'.format(path))

    if present:
        if not host.get_fact(File, path=activate_script_path):
            # Create missing virtualenv
            command = ['virtualenv']

            if venv:
                command = [python or 'python', '-m', 'venv']

            if python and not venv:
                command.append('-p {0}'.format(python))

            if site_packages:
                command.append('--system-site-packages')

            if always_copy and not venv:
                command.append('--always-copy')
            elif always_copy and venv:
                command.append('--copies')

            command.append(path)

            yield ' '.join(command)

            host.create_fact(
                File,
                kwargs={'path': activate_script_path},
                data={'user': None, 'group': None},
            )
        else:
            host.noop('virtualenv {0} exists'.format(path))

_virtualenv = virtualenv  # noqa


@operation
def venv(
    path,
    python=None, site_packages=False, always_copy=False, present=True,
    state=None, host=None,
):
    '''
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    Example:

    .. code:: python

        pip.venv(
            name='Create a virtualenv',
            path='/usr/local/bin/venv',
        )
    '''

    yield virtualenv(
        venv=True,
        path=path,
        python=python,
        site_packages=site_packages,
        always_copy=always_copy,
        present=present,
        state=state,
        host=host,
    )


@operation
def packages(
    packages=None, present=True, latest=False,
    requirements=None, pip='pip', virtualenv=None, virtualenv_kwargs=None,
    extra_install_args=None,
    state=None, host=None,
):
    '''
    Install/remove/update pip packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + requirements: location of requirements file to install/uninstall
    + pip: name or path of the pip directory to use
    + virtualenv: root directory of virtualenv to work in
    + virtualenv_kwargs: dictionary of arguments to pass to ``pip.virtualenv``
    + extra_install_args: additional arguments to the pip install command

    Virtualenv:
        This will be created if it does not exist already. ``virtualenv_kwargs``
        will be passed to ``pip.virtualenv`` which can be used to control how
        the env is created.

    Versions:
        Package versions can be pinned like pip: ``<pkg>==<version>``.

    Example:

    .. code:: python

        pip.packages(
            name='Install pyinfra into a virtualenv',
            packages=['pyinfra'],
            virtualenv='/usr/local/bin/venv',
        )
    '''

    virtualenv_kwargs = virtualenv_kwargs or {}

    # Ensure any virtualenv
    if virtualenv:
        yield _virtualenv(virtualenv, state=state, host=host, **virtualenv_kwargs)

        # And update pip path
        virtualenv = virtualenv.rstrip('/')
        pip = '{0}/bin/{1}'.format(virtualenv, pip)

    install_command = [pip, 'install']
    if extra_install_args:
        install_command.append(extra_install_args)
    install_command = ' '.join(install_command)

    uninstall_command = ' '.join([pip, 'uninstall', '--yes'])

    # (un)Install requirements
    if requirements is not None:
        if present:
            yield '{0} -r {1}'.format(install_command, requirements)
        else:
            yield '{0} -r {1}'.format(uninstall_command, requirements)

    # Handle passed in packages
    if packages:
        current_packages = host.get_fact(PipPackages, pip=pip)

        yield ensure_packages(
            host, packages, current_packages, present,
            install_command=install_command,
            uninstall_command=uninstall_command,
            upgrade_command='{0} --upgrade'.format(install_command),
            version_join='==',
            latest=latest,
        )



'''
Manage dnf packages and repositories. Note that dnf package names are case-sensitive.
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.rpm import RpmPackageProvides, RpmPackages

from . import files
from .util.packaging import ensure_packages, ensure_rpm, ensure_yum_repo


@operation(is_idempotent=False)
def key(src, state=None, host=None):
    '''
    Add dnf gpg keys with ``rpm``.

    + key: filename or URL

    Note:
        always returns one command, not idempotent

    Example:

    .. code:: python

        linux_id = host.get_fact(LinuxDistribution)['release_meta'].get('ID')
        dnf.key(
            name='Add the Docker CentOS gpg key',
            src='https://download.docker.com/linux/{}/gpg'.format(linux_id),
        )

    '''

    yield 'rpm --import {0}'.format(src)


@operation
def repo(
    src,
    present=True,
    baseurl=None, description=None,
    enabled=True, gpgcheck=True, gpgkey=None,
    state=None, host=None,
):
    # NOTE: if updating this docstring also update `yum.repo`
    '''
    Add/remove/update dnf repositories.

    + name: URL or name for the ``.repo``   file
    + present: whether the ``.repo`` file should be present
    + baseurl: the baseurl of the repo (if ``name`` is not a URL)
    + description: optional verbose description
    + enabled: whether this repo is enabled
    + gpgcheck: whether set ``gpgcheck=1``
    + gpgkey: the URL to the gpg key for this repo

    ``Baseurl``/``description``/``gpgcheck``/``gpgkey``:
        These are only valid when ``name`` is a filename (ie not a URL). This is
        for manual construction of repository files. Use a URL to download and
        install remote repository files.

    Examples:

    .. code:: python

        # Download a repository file
        dnf.rpm(
            name='Install Docker-CE repo via URL',
            src='https://download.docker.com/linux/centos/docker-ce.repo',
        )

        # Create the repository file from baseurl/etc
        dnf.repo(
            name='Add the Docker CentOS repo',
            src='DockerCE',
            baseurl='https://download.docker.com/linux/centos/7/$basearch/stable',
        )
    '''

    yield ensure_yum_repo(
        state, host, files,
        src, baseurl, present, description, enabled, gpgcheck, gpgkey,
    )


@operation
def rpm(src, present=True, state=None, host=None):
    # NOTE: if updating this docstring also update `yum.rpm`
    '''
    Add/remove ``.rpm`` file packages.

    + src: filename or URL of the ``.rpm`` package
    + present: whether ore not the package should exist on the system

    URL sources with ``present=False``:
        If the ``.rpm`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    Example:

    .. code:: python

        major_centos_version = host.get_fact(LinuxDistribution)['major']
        dnf.rpm(
           name='Install EPEL rpm to enable EPEL repo',
           src='https://dl.fedoraproject.org/pub/epel/epel-release-latest-{}.noarch.rpm'.format(major_centos_version),
        )
    '''

    yield ensure_rpm(state, host, files, src, present, 'dnf')


@operation
def update(state=None, host=None):
    '''
    Updates all dnf packages.
    '''

    yield 'dnf update -y'

_update = update  # noqa: E305 (for use below where update is a kwarg)


@operation
def packages(
    packages=None,
    present=True, latest=False, update=False, clean=False, nobest=False,
    extra_install_args=None, extra_uninstall_args=None,
    state=None, host=None,
):
    '''
    Install/remove/update dnf packages & updates.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``dnf update`` before installing packages
    + clean: run ``dnf clean`` before installing packages
    + nobest: add the no best option to install
    + extra_install_args: additional arguments to the dnf install command
    + extra_uninstall_args: additional arguments to the dnf uninstall command

    Versions:
        Package versions can be pinned as follows: ``<pkg>=<version>``

    Examples:

    .. code:: python

        # Update package list and install packages
        dnf.packages(
            name='Install Vim and Vim enhanced',
            packages=['vim-enhanced', 'vim'],
            update=True,
        )

        # Install the latest versions of packages (always check)
        dnf.packages(
            name='Install latest Vim',
            packages=['vim'],
            latest=True,
        )
    '''

    if clean:
        yield 'dnf clean all'

    if update:
        yield _update(state=state, host=host)

    install_command = ['dnf', 'install', '-y']

    if nobest:
        install_command.append('--nobest')

    if extra_install_args:
        install_command.append(extra_install_args)

    uninstall_command = ['dnf', 'remove', '-y']

    if extra_uninstall_args:
        uninstall_command.append(extra_uninstall_args)

    yield ensure_packages(
        host, packages, host.get_fact(RpmPackages), present,
        install_command=' '.join(install_command),
        uninstall_command=' '.join(uninstall_command),
        upgrade_command='dnf update -y',
        version_join='=',
        lower=False,  # dnf packages are case sensitive
        latest=latest,
        expand_package_fact=lambda package: host.get_fact(
            RpmPackageProvides,
            name=package,
        ),
    )



'''
Manage BSD init services (``/etc/rc.d``, ``/usr/local/etc/rc.d``).
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.bsdinit import RcdStatus
from pyinfra.facts.server import Os

from . import files
from .util.service import handle_service_control


@operation
def service(
    service,
    running=True, restarted=False, reloaded=False,
    command=None, enabled=None,
    state=None, host=None,
):
    '''
    Manage the state of BSD init services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<service> <command>``
    + enabled: whether this service should be enabled/disabled on boot
    '''

    status_argument = 'status'
    if host.get_fact(Os) == 'OpenBSD':
        status_argument = 'check'

    yield handle_service_control(
        host,
        service, RcdStatus,
        'test -e /etc/rc.d/{0} && /etc/rc.d/{0} {1} || /usr/local/etc/rc.d/{0} {1}',
        running, restarted, reloaded, command,
        status_argument=status_argument,
    )

    # BSD init is simple, just add/remove <service>_enabled="YES"
    if isinstance(enabled, bool):
        yield files.line(
            '/etc/rc.conf.local',
            '^{0}_enable='.format(service),
            replace='{0}_enable="YES"'.format(service),
            present=enabled,
            state=state, host=host,
        )



import re

from pyinfra.api import FactBase


class UpstartStatus(FactBase):
    '''
    Returns a dict of name -> status for upstart managed services.
    '''

    command = 'initctl list'
    requires_command = 'initctl'

    regex = r'^([a-z\-]+) [a-z]+\/([a-z]+)'
    default = dict

    def process(self, output):
        services = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                services[matches.group(1)] = matches.group(2) == 'running'

        return services



import shlex

from string import Formatter

import six

from six.moves import shlex_quote

from .operation_kwargs import get_executor_kwarg_keys


def make_formatted_string_command(string, *args, **kwargs):
    '''
    Helper function that takes a shell command or script as a string, splits it
    using ``shlex.split`` and then formats each bit, returning a ``StringCommand``
    instance with each bit.

    Useful to enable string formatted commands/scripts, for example:

    .. code:: python

        curl_command = make_formatted_string_command(
            'curl -sSLf {0} -o {1}',
            QuoteString(src),
            QuoteString(dest),
        )
    '''

    formatter = Formatter()
    string_bits = []

    for bit in shlex.split(string):
        for item in formatter.parse(bit):
            if item[0]:
                string_bits.append(item[0])
            if item[1]:
                value, _ = formatter.get_field(item[1], args, kwargs)
                string_bits.append(value)

    return StringCommand(*string_bits)


class MaskString(str):
    pass


class QuoteString(object):
    def __init__(self, obj):
        self.object = obj

    def __repr__(self):
        return 'QuoteString({0})'.format(self.object)


class PyinfraCommand(object):
    def __init__(self, *args, **kwargs):
        self.executor_kwargs = {
            key: kwargs[key]
            for key in get_executor_kwarg_keys()
            if key in kwargs
        }

    def __eq__(self, other):
        if isinstance(other, self.__class__) and repr(self) == repr(other):
            return True
        return False

    def execute(self, state, host, executor_kwargs):
        raise NotImplementedError


class StringCommand(PyinfraCommand):
    def __init__(self, *bits, **kwargs):
        super(StringCommand, self).__init__(**kwargs)
        self.bits = bits
        self.separator = kwargs.pop('_separator', ' ')

    def __str__(self):
        return self.get_masked_value()

    def __repr__(self):
        return 'StringCommand({0})'.format(self.get_masked_value())

    def _get_all_bits(self, bit_accessor):
        all_bits = []

        for bit in self.bits:
            quote = False
            if isinstance(bit, QuoteString):
                quote = True
                bit = bit.object

            if isinstance(bit, StringCommand):
                bit = bit_accessor(bit)

            if not isinstance(bit, six.string_types):
                bit = '{0}'.format(bit)

            if quote:
                bit = shlex_quote(bit)

            all_bits.append(bit)

        return all_bits

    def get_raw_value(self):
        return self.separator.join(self._get_all_bits(
            lambda bit: bit.get_raw_value(),
        ))

    def get_masked_value(self):
        return self.separator.join([
            '***' if isinstance(bit, MaskString) else bit
            for bit in self._get_all_bits(lambda bit: bit.get_masked_value())
        ])

    def execute(self, state, host, executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.run_shell_command(
            self,
            print_output=state.print_output,
            print_input=state.print_input,
            return_combined_output=True,
            **executor_kwargs
        )


class FileUploadCommand(PyinfraCommand):
    def __init__(self, src, dest, **kwargs):
        super(FileUploadCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest

    def __repr__(self):
        return 'FileUploadCommand({0}, {1})'.format(self.src, self.dest)

    def execute(self, state, host, executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.put_file(
            self.src, self.dest,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs
        )


class FileDownloadCommand(PyinfraCommand):
    def __init__(self, src, dest, **kwargs):
        super(FileDownloadCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest

    def __repr__(self):
        return 'FileDownloadCommand({0}, {1})'.format(self.src, self.dest)

    def execute(self, state, host, executor_kwargs):
        executor_kwargs.update(self.executor_kwargs)

        return host.get_file(
            self.src, self.dest,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs
        )


class FunctionCommand(PyinfraCommand):
    def __init__(self, function, args, func_kwargs, **kwargs):
        super(FunctionCommand, self).__init__(**kwargs)
        self.function = function
        self.args = args
        self.kwargs = func_kwargs

    def __repr__(self):
        return 'FunctionCommand({0}, {1}, {2})'.format(
            self.function.__name__,
            self.args,
            self.kwargs,
        )

    def execute(self, state, host, executor_kwargs):
        return self.function(state, host, *self.args, **self.kwargs)


class RsyncCommand(PyinfraCommand):
    def __init__(self, src, dest, flags, **kwargs):
        super(RsyncCommand, self).__init__(**kwargs)
        self.src = src
        self.dest = dest
        self.flags = flags

    def __repr__(self):
        return 'RsyncCommand({0}, {1}, {2})'.format(self.src, self.dest, self.flags)

    def execute(self, state, host, executor_kwargs):
        return host.rsync(
            self.src, self.dest, self.flags,
            print_output=state.print_output,
            print_input=state.print_input,
            **executor_kwargs
        )



from __future__ import unicode_literals

import re

from datetime import datetime

from pyinfra.api.command import make_formatted_string_command, QuoteString
from pyinfra.api.facts import FactBase
from pyinfra.api.util import try_int

LINUX_STAT_COMMAND = (
    "stat -c 'user=%U group=%G mode=%A atime=%X mtime=%Y ctime=%Z size=%s %N'"
)
BSD_STAT_COMMAND = (
    "stat -f 'user=%Su group=%Sg mode=%Sp atime=%a mtime=%m ctime=%c size=%z %N%SY'"
)

STAT_REGEX = (
    r'user=(.*) group=(.*) mode=(.*) '
    r'atime=([0-9]*) mtime=([0-9]*) ctime=([0-9]*) '
    r'size=([0-9]*) (.*)'
)

FLAG_TO_TYPE = {
    'b': 'block',
    'c': 'character',
    'd': 'directory',
    'l': 'link',
    's': 'socket',
    'p': 'fifo',
    '-': 'file',
}

SYMBOL_TO_OCTAL_PERMISSIONS = {
    'rwx': '7',
    'rw-': '6',
    'r-x': '5',
    'r--': '4',
    '-wx': '3',
    '-w-': '2',
    '--x': '1',
}


def _parse_mode(mode):
    '''
    Converts ls mode output (rwxrwxrwx) -> integer (755).
    '''

    result = ''
    # owner, group, world
    for group in [mode[0:3], mode[3:6], mode[6:9]]:
        if group in SYMBOL_TO_OCTAL_PERMISSIONS:
            result = '{0}{1}'.format(result, SYMBOL_TO_OCTAL_PERMISSIONS[group])
        else:
            result = '{0}0'.format(result)

    return int(result)


class File(FactBase):
    # Types must match FLAG_TO_TYPE in .util.files.py
    type = 'file'
    test_flag = '-e'

    def command(self, path):
        return make_formatted_string_command((
            '! test {test_flag} {0} || '  # only stat if the file exists
            '( {linux_stat_command} {0} 2> /dev/null || {bsd_stat_command} {0} )'
        ),
            QuoteString(path),
            test_flag=self.test_flag,
            linux_stat_command=LINUX_STAT_COMMAND,
            bsd_stat_command=BSD_STAT_COMMAND,
        )

    def process(self, output):
        match = re.match(STAT_REGEX, output[0])
        if not match:
            return None

        data = {}
        path_type = None

        for key, value in (
            ('user', match.group(1)),
            ('group', match.group(2)),
            ('mode', match.group(3)),
            ('atime', match.group(4)),
            ('mtime', match.group(5)),
            ('ctime', match.group(6)),
            ('size', match.group(7)),
        ):
            if key == 'mode':
                path_type = FLAG_TO_TYPE[value[0]]
                value = _parse_mode(value[1:])

            elif key == 'size':
                value = try_int(value)

            elif key in ('atime', 'mtime', 'ctime'):
                value = try_int(value)
                if isinstance(value, int):
                    value = datetime.utcfromtimestamp(value)

            data[key] = value

        if path_type != self.type:
            return False

        if path_type == 'link':
            filename = match.group(8)
            filename, target = filename.split(' -> ')
            data['link_target'] = target.strip("'")

        return data


class Link(File):
    type = 'link'
    test_flag = '-L'


class Directory(File):
    type = 'directory'


class Socket(File):
    type = 'socket'


class Sha1File(FactBase):
    '''
    Returns a SHA1 hash of a file. Works with both sha1sum and sha1.
    '''

    _regexes = [
        r'^([a-zA-Z0-9]{40})\s+%s$',
        r'^SHA1\s+\(%s\)\s+=\s+([a-zA-Z0-9]{40})$',
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command((
            'test -e {0} && ( '
            'sha1sum {0} 2> /dev/null || shasum {0} 2> /dev/null || sha1 {0} '
            ') || true'
        ), QuoteString(path))

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class Sha256File(FactBase):
    '''
    Returns a SHA256 hash of a file.
    '''

    _regexes = [
        r'^([a-zA-Z0-9]{64})\s+%s$',
        r'^SHA256\s+\(%s\)\s+=\s+([a-zA-Z0-9]{64})$',
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command((
            'test -e {0} && ( '
            'sha256sum {0} 2> /dev/null '
            '|| shasum -a 256 {0} 2> /dev/null '
            '|| sha256 {0} '
            ') || true'
        ), QuoteString(path))

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class Md5File(FactBase):
    '''
    Returns an MD5 hash of a file.
    '''

    _regexes = [
        r'^([a-zA-Z0-9]{32})\s+%s$',
        r'^SHA256\s+\(%s\)\s+=\s+([a-zA-Z0-9]{32})$',
    ]

    def command(self, path):
        self.path = path
        return make_formatted_string_command(
            'test -e {0} && ( md5sum {0} 2> /dev/null || md5 {0} ) || true',
            QuoteString(path),
        )

    def process(self, output):
        for regex in self._regexes:
            regex = regex % re.escape(self.path)
            matches = re.match(regex, output[0])
            if matches:
                return matches.group(1)


class FindInFile(FactBase):
    '''
    Checks for the existence of text in a file using grep. Returns a list of matching
    lines if the file exists, and ``None`` if the file does not.
    '''

    def command(self, path, pattern):
        self.exists_flag = '__pyinfra_exists_{0}'.format(path)

        return make_formatted_string_command((
            'grep -e {0} {1} 2> /dev/null || '
            '( find {1} -type f > /dev/null && echo {2} || true )'
        ), QuoteString(pattern), QuoteString(path), QuoteString(self.exists_flag))

    def process(self, output):
        # If output is the special string: no matches, so return an empty list;
        # this allows us to differentiate between no matches in an existing file
        # or a file not existing.
        if output and output[0] == self.exists_flag:
            return []

        return output


class FindFiles(FactBase):
    '''
    Returns a list of files from a start point, recursively using find.
    '''

    type_flag = 'f'

    def command(self, path, quote_path=True):
        return make_formatted_string_command(
            'find {0} -type {type_flag} || true',
            QuoteString(path) if quote_path else path,
            type_flag=self.type_flag,
        )

    @staticmethod
    def process(output):
        return output


class FindLinks(FindFiles):
    '''
    Returns a list of links from a start point, recursively using find.
    '''

    type_flag = 'l'


class FindDirectories(FindFiles):
    '''
    Returns a list of directories from a start point, recursively using find.
    '''

    type_flag = 'd'



from collections import defaultdict
from inspect import currentframe, getframeinfo
from os import path
from unittest import TestCase

from mock import mock_open, patch

import pyinfra

from pyinfra import pseudo_host, pseudo_state
from pyinfra.api import (
    BaseStateCallback,
    Config,
    FileDownloadCommand,
    FileUploadCommand,
    OperationError,
    OperationValueError,
    State,
    StringCommand,
)
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.operation import add_op, get_operation_names, OperationMeta
from pyinfra.api.operations import run_ops
from pyinfra.operations import files, python, server

from ..paramiko_util import (
    FakeBuffer,
    FakeChannel,
    PatchSSHTestCase,
)
from ..util import FakeState, make_inventory


class TestOperationMeta(TestCase):
    def test_get_operation_names(self):
        assert len(get_operation_names()) > 0

    def test_operation_meta_repr(self):
        op_meta = OperationMeta('hash', [])
        assert repr(op_meta) == 'commands:[] changed:False hash:hash'


class TestOperationsApi(PatchSSHTestCase):
    def test_op(self):
        inventory = make_inventory()
        somehost = inventory.get_host('somehost')
        anotherhost = inventory.get_host('anotherhost')

        state = State(inventory, Config())
        state.add_callback_handler(BaseStateCallback())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        add_op(
            state, files.file,
            '/var/log/pyinfra.log',
            user='pyinfra',
            group='pyinfra',
            mode='644',
            create_remote_dir=False,
            sudo=True,
            sudo_user='test_sudo',
            su_user='test_su',
            ignore_errors=True,
            env={
                'TEST': 'what',
            },
        )

        op_order = state.get_op_order()

        # Ensure we have an op
        assert len(op_order) == 1

        first_op_hash = op_order[0]

        # Ensure the op name
        assert state.op_meta[first_op_hash]['names'] == {'Files/File'}

        # Ensure the commands
        assert state.ops[somehost][first_op_hash]['commands'] == [
            StringCommand('touch /var/log/pyinfra.log'),
            StringCommand('chmod 644 /var/log/pyinfra.log'),
            StringCommand('chown pyinfra:pyinfra /var/log/pyinfra.log'),
        ]

        # Ensure the global kwargs (same for both hosts)
        somehost_global_kwargs = state.ops[somehost][first_op_hash]['global_kwargs']
        assert somehost_global_kwargs['sudo'] is True
        assert somehost_global_kwargs['sudo_user'] == 'test_sudo'
        assert somehost_global_kwargs['su_user'] == 'test_su'
        assert somehost_global_kwargs['ignore_errors'] is True

        anotherhost_global_kwargs = state.ops[anotherhost][first_op_hash]['global_kwargs']
        assert anotherhost_global_kwargs['sudo'] is True
        assert anotherhost_global_kwargs['sudo_user'] == 'test_sudo'
        assert anotherhost_global_kwargs['su_user'] == 'test_su'
        assert anotherhost_global_kwargs['ignore_errors'] is True

        # Ensure run ops works
        run_ops(state)

        # Ensure ops completed OK
        assert state.results[somehost]['success_ops'] == 1
        assert state.results[somehost]['ops'] == 1
        assert state.results[anotherhost]['success_ops'] == 1
        assert state.results[anotherhost]['ops'] == 1

        # And w/o errors
        assert state.results[somehost]['error_ops'] == 0
        assert state.results[anotherhost]['error_ops'] == 0

        # And with the different modes
        run_ops(state, serial=True)
        run_ops(state, no_wait=True)

        disconnect_all(state)

    def test_op_call_direct_falls(self):
        inventory = make_inventory()
        somehost = inventory.get_host('somehost')
        state = State(inventory, Config())

        # Enable printing on this test to catch any exceptions in the formatting
        state.print_output = True
        state.print_input = True
        state.print_fact_info = True
        state.print_noop_info = True

        connect_all(state)

        with self.assertRaises(PyinfraError) as context:
            server.shell(state=state, host=somehost, commands='echo hi')

        assert context.exception.args[0] == (
            'Operation order number not provided in API mode - '
            'you must use `add_op` to add operations.'
        )

    def test_file_upload_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        connect_all(state)

        with patch('pyinfra.operations.files.os_path.isfile', lambda *args, **kwargs: True):
            # Test normal
            add_op(
                state, files.put,
                name='First op name',
                src='files/file.txt',
                dest='/home/vagrant/file.txt',
            )

            # And with sudo
            add_op(
                state, files.put,
                src='files/file.txt',
                dest='/home/vagrant/file.txt',
                sudo=True,
                sudo_user='pyinfra',
            )

            # And with su
            add_op(
                state, files.put,
                src='files/file.txt',
                dest='/home/vagrant/file.txt',
                sudo=True,
                su_user='pyinfra',
            )

        op_order = state.get_op_order()

        # Ensure we have all ops
        assert len(op_order) == 3

        first_op_hash = op_order[0]
        second_op_hash = op_order[1]

        # Ensure first op is the right one
        assert state.op_meta[first_op_hash]['names'] == {'First op name'}

        somehost = inventory.get_host('somehost')
        anotherhost = inventory.get_host('anotherhost')

        # Ensure first op has the right (upload) command
        assert state.ops[somehost][first_op_hash]['commands'] == [
            StringCommand('mkdir -p /home/vagrant'),
            FileUploadCommand('files/file.txt', '/home/vagrant/file.txt'),
        ]

        # Ensure second op has sudo/sudo_user
        assert state.ops[somehost][second_op_hash]['global_kwargs']['sudo'] is True
        assert state.ops[somehost][second_op_hash]['global_kwargs']['sudo_user'] == 'pyinfra'

        # Ensure third has su_user
        assert state.ops[somehost][op_order[2]]['global_kwargs']['su_user'] == 'pyinfra'

        # Check run ops works
        with patch('pyinfra.api.util.open', mock_open(read_data='test!'), create=True):
            run_ops(state)

        # Ensure ops completed OK
        assert state.results[somehost]['success_ops'] == 3
        assert state.results[somehost]['ops'] == 3
        assert state.results[anotherhost]['success_ops'] == 3
        assert state.results[anotherhost]['ops'] == 3

        # And w/o errors
        assert state.results[somehost]['error_ops'] == 0
        assert state.results[anotherhost]['error_ops'] == 0

    def test_file_download_op(self):
        inventory = make_inventory()

        state = State(inventory, Config())
        connect_all(state)

        with patch('pyinfra.operations.files.os_path.isfile', lambda *args, **kwargs: True):
            add_op(
                state, files.get,
                name='First op name',
                src='/home/vagrant/file.txt',
                dest='files/file.txt',
            )

        op_order = state.get_op_order()

        assert len(op_order) == 1

        first_op_hash = op_order[0]
        assert state.op_meta[first_op_hash]['names'] == {'First op name'}

        somehost = inventory.get_host('somehost')
        anotherhost = inventory.get_host('anotherhost')

        # Ensure first op has the right (upload) command
        assert state.ops[somehost][first_op_hash]['commands'] == [
            FileDownloadCommand('/home/vagrant/file.txt', 'files/file.txt'),
        ]

        with patch('pyinfra.api.util.open', mock_open(read_data='test!'), create=True):
            run_ops(state)

        assert state.results[somehost]['success_ops'] == 1
        assert state.results[somehost]['ops'] == 1
        assert state.results[anotherhost]['success_ops'] == 1
        assert state.results[anotherhost]['ops'] == 1
        assert state.results[somehost]['error_ops'] == 0
        assert state.results[anotherhost]['error_ops'] == 0

    def test_function_call_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        is_called = []

        def mocked_function(*args, **kwargs):
            is_called.append(True)
            return None

        # Add op to both hosts
        add_op(state, python.call, mocked_function)

        # Ensure there is one op
        assert len(state.get_op_order()) == 1

        run_ops(state)

        assert is_called

    def test_run_once_serial_op(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        # Add a run once op
        add_op(state, server.shell, 'echo "hi"', run_once=True, serial=True)

        # Ensure it's added to op_order
        assert len(state.get_op_order()) == 1

        somehost = inventory.get_host('somehost')
        anotherhost = inventory.get_host('anotherhost')

        # Ensure between the two hosts we only run the one op
        assert len(state.ops[somehost]) + len(state.ops[anotherhost]) == 1

        # Check run works
        run_ops(state)

        assert (
            state.results[somehost]['success_ops']
            + state.results[anotherhost]['success_ops']
        ) == 1

    def test_rsync_op(self):
        inventory = make_inventory(hosts=('somehost',))
        state = State(inventory, Config())
        connect_all(state)

        with patch('pyinfra.api.connectors.ssh.check_can_rsync'):
            add_op(state, files.rsync, 'src', 'dest', sudo=True, sudo_user='root')

        assert len(state.get_op_order()) == 1

        with patch('pyinfra.api.connectors.ssh.run_local_process') as fake_run_local_process:
            fake_run_local_process.return_value = 0, []
            run_ops(state)

        fake_run_local_process.assert_called_with(
            (
                'rsync -ax --delete --rsh '
                "'ssh -o BatchMode=yes -o StrictHostKeyChecking=no '"
                " --rsync-path 'sudo -u root rsync' src vagrant@somehost:dest"
            ),
            print_output=False,
            print_prefix=inventory.get_host('somehost').print_prefix,
        )

    def test_rsync_op_failure(self):
        inventory = make_inventory(hosts=('somehost',))
        state = State(inventory, Config())
        connect_all(state)

        with patch('pyinfra.api.connectors.ssh.find_executable', lambda x: None):
            with self.assertRaises(OperationError) as context:
                add_op(state, files.rsync, 'src', 'dest')

        assert context.exception.args[0] == 'The `rsync` binary is not available on this system.'

    def test_op_cannot_change_execution_kwargs(self):
        inventory = make_inventory()

        state = State(inventory, Config())

        class NoSetDefaultDict(defaultdict):
            def setdefault(self, key, _):
                return self[key]

        state.op_meta = NoSetDefaultDict(lambda: {'serial': True})

        connect_all(state)

        with self.assertRaises(OperationValueError) as context:
            add_op(state, files.file, '/var/log/pyinfra.log', serial=False)

        assert context.exception.args[0] == 'Cannot have different values for `serial`.'


class TestOperationFailures(PatchSSHTestCase):
    def test_full_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"')

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer('', fake_channel),
            )

            with self.assertRaises(PyinfraError) as e:
                run_ops(state)

            assert e.exception.args[0] == 'No hosts remaining!'

            somehost = inventory.get_host('somehost')

            # Ensure the op was not flagged as success
            assert state.results[somehost]['success_ops'] == 0
            # And was flagged asn an error
            assert state.results[somehost]['error_ops'] == 1

    def test_ignore_errors_op_fail(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        add_op(state, server.shell, 'echo "hi"', ignore_errors=True)

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_channel = FakeChannel(1)
            fake_run_command.return_value = (
                False,
                FakeBuffer('', fake_channel),
            )

            # This should run OK
            run_ops(state)

        somehost = inventory.get_host('somehost')

        # Ensure the op was added to results
        assert state.results[somehost]['ops'] == 1
        assert state.results[somehost]['error_ops'] == 1
        # But not as a success
        assert state.results[somehost]['success_ops'] == 0

    def test_no_invalid_op_call(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)
        pseudo_state.set(state)

        state.in_op = True
        with self.assertRaises(PyinfraError):
            server.user('someuser')

        state.in_op = False
        state.in_deploy = True
        with self.assertRaises(PyinfraError):
            server.user('someuser')


class TestOperationOrdering(PatchSSHTestCase):
    # In CLI mode, pyinfra uses *line numbers* to order operations as defined by
    # the user. This makes reasoning about user-written deploys simple and easy
    # to understand.
    def test_cli_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        state.current_deploy_filename = __file__

        pyinfra.is_cli = True
        pseudo_state.set(state)

        # Add op to both hosts
        for name in ('anotherhost', 'somehost'):
            pseudo_host.set(inventory.get_host(name))
            server.shell('echo hi')  # note this is called twice but on *the same line*

        # Add op to just the second host - using the pseudo modules such that
        # it replicates a deploy file.
        pseudo_host.set(inventory.get_host('anotherhost'))
        first_pseudo_hash = server.user('anotherhost_user').hash
        first_pseudo_call_line = getframeinfo(currentframe()).lineno - 1

        # Add op to just the first host - using the pseudo modules such that
        # it replicates a deploy file.
        pseudo_host.set(inventory.get_host('somehost'))
        second_pseudo_hash = server.user('somehost_user').hash
        second_pseudo_call_line = getframeinfo(currentframe()).lineno - 1

        pseudo_state.reset()
        pseudo_host.reset()

        pyinfra.is_cli = False

        # Ensure there are two ops
        op_order = state.get_op_order()
        assert len(op_order) == 3

        # And that the two ops above were called in the expected order
        assert op_order[1] == first_pseudo_hash
        assert op_order[2] == second_pseudo_hash

        # And that they have the expected line numbers
        assert state.op_line_numbers_to_hash.get((first_pseudo_call_line,)) == first_pseudo_hash
        assert state.op_line_numbers_to_hash.get((second_pseudo_call_line,)) == second_pseudo_hash

        # Ensure somehost has two ops and anotherhost only has the one
        assert len(state.ops[inventory.get_host('somehost')]) == 2
        assert len(state.ops[inventory.get_host('anotherhost')]) == 2

    # In API mode, pyinfra *overrides* the line numbers such that whenever an
    # operation or deploy is added it is simply appended. This makes sense as
    # the user writing the API calls has full control over execution order.
    def test_api_op_line_numbers(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)

        another_host = inventory.get_host('anotherhost')

        def add_another_op():
            return add_op(state, server.shell, 'echo second-op')[another_host].hash

        first_op_hash = add_op(state, server.shell, 'echo first-op')[another_host].hash
        second_op_hash = add_another_op()  # note `add_op` will be called on an earlier line

        op_order = state.get_op_order()
        assert len(op_order) == 2

        assert op_order[0] == first_op_hash
        assert op_order[1] == second_op_hash


this_filename = path.join('tests', 'test_api', 'test_api_operations.py')


class TestOperationExceptions(TestCase):
    def test_add_op_rejects_cli(self):
        pyinfra.is_cli = True

        with self.assertRaises(PyinfraError) as context:
            add_op(None, server.shell)
        call_line = getframeinfo(currentframe()).lineno - 1

        pyinfra.is_cli = False

        assert context.exception.args[0] == (
            '`add_op` should not be called when pyinfra is executing in CLI mode! '
            '(line {0} in {1})'.format(call_line, this_filename)
        )

    def test_op_call_rejects_no_cli(self):
        with self.assertRaises(PyinfraError) as context:
            server.shell()
        call_line = getframeinfo(currentframe()).lineno - 1

        assert context.exception.args[0] == (
            'API operation called without state/host: '
            'server.shell (line {0} in {1})'.format(call_line, this_filename)
        )

    def test_op_call_rejects_in_op(self):
        state = FakeState()

        pyinfra.is_cli = True
        pseudo_state.set(state)

        with self.assertRaises(PyinfraError) as context:
            server.shell()
        call_line = getframeinfo(currentframe()).lineno - 1

        pyinfra.is_cli = False
        pseudo_state.reset()

        assert context.exception.args[0] == (
            'Nested operation called without state/host: '
            'server.shell (line {0} in {1})'.format(call_line, this_filename)
        )

    def test_op_call_rejects_in_deploy(self):
        state = FakeState()
        state.in_op = False

        pyinfra.is_cli = True
        pseudo_state.set(state)

        with self.assertRaises(PyinfraError) as context:
            server.shell()
        call_line = getframeinfo(currentframe()).lineno - 1

        pyinfra.is_cli = False
        pseudo_state.reset()

        assert context.exception.args[0] == (
            'Nested deploy operation called without state/host: '
            'server.shell (line {0} in {1})'.format(call_line, this_filename)
        )



import json

from pyinfra.api import FactBase


class OpenvzContainers(FactBase):
    '''
    Returns a dict of running OpenVZ containers by CTID:

    .. code:: python

        {
            666: {
                'ip': [],
                'ostemplate': 'ubuntu...',
                ...
            },
        }
    '''

    command = 'vzlist -a -j'
    requires_command = 'vzlist'

    default = dict

    @staticmethod
    def process(output):
        combined_json = ''.join(output)
        vz_data = json.loads(combined_json)

        return {
            int(vz.pop('ctid')): vz
            for vz in vz_data
        }



'''
The windows_files module handles windows filesystem state, file uploads and template generation.
'''

from __future__ import unicode_literals

import ntpath
import os
from datetime import timedelta

import six

from pyinfra import logger
from pyinfra.api import (
    FileUploadCommand,
    operation,
    OperationError,
    OperationTypeError,
)
from pyinfra.api.util import get_file_sha1

from .util.compat import fspath
from .util.files import ensure_mode_int


@operation(pipeline_facts={
    'file': 'dest',
})
def download(
    src, dest,
    user=None, group=None, mode=None, cache_time=None, force=False,
    sha256sum=None, sha1sum=None, md5sum=None,
    state=None, host=None,
):
    '''
    Download files from remote locations using curl or wget.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against

    Example:

    .. code:: python

        winows_files.download(
            name='Download the Docker repo file',
            src='https://download.docker.com/linux/centos/docker-ce.repo',
            dest='C:\\docker',
        )
    '''

    info = host.fact.windows_file(dest)
    # Destination is a directory?
    if info is False:
        raise OperationError(
            'Destination {0} already exists and is not a file'.format(dest),
        )

    # Do we download the file? Force by default
    download = force

    # Doesn't exist, lets download it
    if info is None:
        download = True

    # Destination file exists & cache_time: check when the file was last modified,
    # download if old
    else:
        if cache_time:
            # Time on files is not tz-aware, and will be the same tz as the server's time,
            # so we can safely remove the tzinfo from host.fact.date before comparison.
            cache_time = host.fact.windows_date.replace(tzinfo=None) - timedelta(seconds=cache_time)
            if info['mtime'] and info['mtime'] > cache_time:
                download = True

        if sha1sum:
            if sha1sum != host.fact.windows_sha1_file(dest):
                download = True

        if sha256sum:
            if sha256sum != host.fact.windows_sha256_file(dest):
                download = True

        if md5sum:
            if md5sum != host.fact.windows_md5_file(dest):
                download = True

    # If we download, always do user/group/mode as SSH user may be different
    if download:
        yield (
            '$ProgressPreference = "SilentlyContinue"; '
            'Invoke-WebRequest -Uri {0} -OutFile {1}'
        ).format(src, dest)

        # if user or group:
        #    yield chown(dest, user, group)

        # if mode:
        #    yield chmod(dest, mode)

        if sha1sum:
            yield (
                'if ((Get-FileHash -Algorithm SHA1 "{0}").hash -ne {1}) {{ '
                'Write-Error "SHA1 did not match!" '
                '}}'
            ).format(dest, sha1sum)

        if sha256sum:
            yield (
                'if ((Get-FileHash -Algorithm SHA256 "{0}").hash -ne {1}) {{ '
                'Write-Error "SHA256 did not match!" '
                '}}'
            ).format(dest, sha256sum)

        if md5sum:
            yield (
                'if ((Get-FileHash -Algorithm MD5 "{0}").hash -ne {1}) {{ '
                'Write-Error "MD5 did not match!" '
                '}}'
            ).format(dest, md5sum)

    else:
        host.noop('file {0} has already been downloaded'.format(dest))


@operation(pipeline_facts={
    'file': 'dest',
    'sha1_file': 'dest',
})
def put(
    src, dest,
    user=None, group=None, mode=None, add_deploy_dir=True,
    create_remote_dir=True, force=False, assume_exists=False,
    state=None, host=None,
):
    '''
    Upload a local file to the remote system.

    + src: local filename to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Note:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

    Examples:

    .. code:: python

        # Note: This requires a 'files/motd' file on the local filesystem
        files.put(
            name='Update the message of the day file',
            src='data/content.json',
            dest='C:\\data\\content.json'
        )
    '''

    # Upload IO objects as-is
    if hasattr(src, 'read'):
        local_file = src

    # Assume string filename
    else:
        # Add deploy directory?
        if add_deploy_dir and state.deploy_dir:
            src = os.path.join(state.deploy_dir, src)

        local_file = src

        if not assume_exists and not os.path.isfile(local_file):
            raise IOError('No such file: {0}'.format(local_file))

    mode = ensure_mode_int(mode)
    remote_file = host.fact.windows_file(dest)

    if create_remote_dir:
        yield _create_remote_dir(state, host, dest, user, group)

    # No remote file, always upload and user/group/mode if supplied
    if not remote_file or force:
        yield FileUploadCommand(local_file, dest)

        # if user or group:
        #    yield chown(dest, user, group)

        # if mode:
        #    yield chmod(dest, mode)

    # File exists, check sum and check user/group/mode if supplied
    else:
        local_sum = get_file_sha1(src)
        remote_sum = host.fact.windows_sha1_file(dest)

        # Check sha1sum, upload if needed
        if local_sum != remote_sum:
            yield FileUploadCommand(local_file, dest)

            # if user or group:
            #    yield chown(dest, user, group)

            # if mode:
            #    yield chmod(dest, mode)

        else:
            changed = False

            # Check mode
            # if mode and remote_file['mode'] != mode:
            #    yield chmod(dest, mode)
            #    changed = True

            # Check user/group
            # if (
            #    (user and remote_file['user'] != user)
            #    or (group and remote_file['group'] != group)
            # ):
            #    yield chown(dest, user, group)
            #    changed = True

            if not changed:
                host.noop('file {0} is already uploaded'.format(dest))


@operation(pipeline_facts={
    'windows_file': 'name',
})
def file(
    path,
    present=True, assume_present=False,
    user=None, group=None, mode=None, touch=False,
    create_remote_dir=True,
    state=None, host=None,
):
    '''
    Add/remove/update files.

    + path: path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + TODO: user: user to own the files
    + TODO: group: group to own the files
    + TODO: mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Example:

    .. code:: python

        files.file(
            name='Create c:\\temp\\hello.txt',
            path='c:\\temp\\hello.txt',
            touch=True,
        )
    '''

    if not isinstance(path, six.string_types):
        raise OperationTypeError('Name must be a string')

    # mode = ensure_mode_int(mode)
    info = host.fact.windows_file(path)

    # Not a file?!
    if info is False:
        raise OperationError('{0} exists and is not a file'.format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        if create_remote_dir:
            yield _create_remote_dir(state, host, path, user, group)

        yield 'New-Item -ItemType file {0}'.format(path)

#        if mode:
#            yield chmod(path, mode)
#        if user or group:
#            yield chown(path, user, group)

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield 'Remove-Item {0}'.format(path)

#    # It exists & we want to ensure its state
#    elif (assume_present or info) and present:
#        if touch:
#            yield 'New-Item -ItemType file {0}'.format(path)
#
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(path, mode)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(path, user, group)


def windows_file(*args, **kwargs):
    # COMPAT
    # TODO: remove this
    logger.warning((
        'Use of `windows_files.windows_file` is deprecated, '
        'please use `windows_files.file` instead.'
    ))
    return file(*args, **kwargs)


def _create_remote_dir(state, host, remote_filename, user, group):
    # Always use POSIX style path as local might be Windows, remote always *nix
    remote_dirname = ntpath.dirname(remote_filename)
    if remote_dirname:
        yield directory(
            remote_dirname,
            state=state,
            host=host,
            user=user,
            group=group,
        )


@operation(pipeline_facts={
    'windows_directory': 'name',
})
def directory(
    path,
    present=True, assume_present=False,
    user=None, group=None, mode=None, recursive=False,
    state=None, host=None,
):
    '''
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + TODO: user: user to own the folder
    + TODO: group: group to own the folder
    + TODO: mode: permissions of the folder
    + TODO: recursive: recursively apply user/group/mode

    Examples:

    .. code:: python

        files.directory(
            name='Ensure the c:\\temp\\dir_that_we_want_removed is removed',
            path='c:\\temp\\dir_that_we_want_removed',
            present=False,
        )

        files.directory(
            name='Ensure c:\\temp\\foo\\foo_dir exists',
            path='c:\\temp\\foo\\foo_dir',
            recursive=True,
        )

        # multiple directories
        dirs = ['c:\\temp\\foo_dir1', 'c:\\temp\\foo_dir2']
        for dir in dirs:
            files.directory(
                name='Ensure the directory `{}` exists'.format(dir),
                path=dir,
            )

    '''

    if not isinstance(path, six.string_types):
        raise OperationTypeError('Name must be a string')

    info = host.fact.windows_directory(path)

    # Not a directory?!
    if info is False:
        raise OperationError('{0} exists and is not a directory'.format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        yield 'New-Item -Path {0} -ItemType Directory'.format(path)
#        if mode:
#            yield chmod(path, mode, recursive=recursive)
#        if user or group:
#            yield chown(path, user, group, recursive=recursive)
#
        # Somewhat bare fact, should flesh out more
        host.fact._create(
            'windows_directory',
            args=(path,),
            data={'type': 'directory'},
        )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        # TODO: how to ensure we use 'ps'?
        # remove anything in the directory
        yield 'Get-ChildItem {0} -Recurse | Remove-Item'.format(path)
        # remove directory
        yield 'Remove-Item {0}'.format(path)

    # It exists & we want to ensure its state
#    elif (assume_present or info) and present:
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(path, mode, recursive=recursive)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(path, user, group, recursive=recursive)


def windows_directory(*args, **kwargs):
    # COMPAT
    # TODO: remove this
    logger.warning((
        'Use of `windows_files.windows_directory` is deprecated, '
        'please use `windows_files.directory` instead.'
    ))
    return directory(*args, **kwargs)


def _validate_path(path):
    try:
        path = fspath(path)
    except TypeError:
        raise OperationTypeError('`path` must be a string or `os.PathLike` object')


@operation(pipeline_facts={
    'link': 'path',
})
def link(
    path,
    target=None, present=True, assume_present=False,
    user=None, group=None, symbolic=True, force=True,
    create_remote_dir=True,
    state=None, host=None,
):
    '''
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + assume_present: whether to assume the link exists
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    Examples:

    .. code:: python

        # simple example showing how to link to a file
        files.link(
            name=r'Create link C:\\issue2 that points to C:\\issue',
            path=r'C:\\issue2',
            target=r'C\\issue',
        )
    '''

    _validate_path(path)

    if present and not target:
        raise OperationError('If present is True target must be provided')

    info = host.fact.windows_link(path)

    # Not a link?
    if info is not None and not info:
        raise OperationError('{0} exists and is not a link'.format(path))

    add_cmd = 'New-Item -ItemType {0} -Path {1} -Target {2} {3}'.format(
        'SymbolicLink' if symbolic else 'HardLink',
        path,
        target,
        '-Force' if force else '',
    )

    remove_cmd = '(Get-Item {0}).Delete()'.format(path)

    # We will attempt to link regardless of current existence
    # since we know by now the path is either a link already
    # or does not exist
    if (info is None or force) and present:
        if create_remote_dir:
            yield _create_remote_dir(state, host, path, user, group)

        yield add_cmd

        # if user or group:
        #    yield chown(path, user, group, dereference=False)

        # host.fact._create(
        #    'windows_link',
        #    args=(path,),
        #    data={'link_target': target, 'group': group, 'user': user},
        # )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield remove_cmd
        # host.fact._delete('windows_link', args=(path,))

    else:
        host.noop('link {0} already exists and force=False'.format(path))



'''
This file as originally part of the "sshuserclient" pypi package. The GitHub
source has now vanished (https://github.com/tobald/sshuserclient).
'''

import glob
import re

from os import path

from paramiko import SSHConfig as ParamikoSSHConfig
from six import StringIO

SETTINGS_REGEX = re.compile(r'(\w+)(?:\s*=\s*|\s+)(.+)')


def _expand_include_statements(file_obj, parsed_files=None):
    parsed_lines = []

    for line in file_obj:
        line = line.strip()
        if not line or line.startswith('#'):
            continue

        match = re.match(SETTINGS_REGEX, line)
        if not match:
            parsed_lines.append(line)
            continue

        key = match.group(1).lower()
        value = match.group(2)

        if key != 'include':
            parsed_lines.append(line)
            continue

        if parsed_files is None:
            parsed_files = []

        # The path can be relative to its parent configuration file
        if path.isabs(value) is False and value[0] != '~':
            folder = path.dirname(file_obj.name)
            value = path.join(folder, value)

        value = path.expanduser(value)

        for filename in glob.iglob(value):
            if path.isfile(filename):
                if filename in parsed_files:
                    raise Exception(
                        'Include loop detected in ssh config file: %s' % filename,
                    )
                with open(filename) as fd:
                    parsed_files.append(filename)
                    parsed_lines.extend(_expand_include_statements(fd, parsed_files))

    return parsed_lines
    output = StringIO('\n'.join(parsed_lines))
    output.name = file_obj.name
    return output


class SSHConfig(ParamikoSSHConfig):
    '''
    an SSHConfig that supports includes directives
    https://github.com/paramiko/paramiko/pull/1194
    '''

    def parse(self, file_obj):
        file_obj = _expand_include_statements(file_obj)
        return super(SSHConfig, self).parse(file_obj)



'''
Manage BSD packages and repositories. Note that BSD package names are case-sensitive.
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.files import File
from pyinfra.facts.pkg import PkgPackages
from pyinfra.facts.server import Arch, Os, OsVersion, Which

from .util.packaging import ensure_packages


@operation
def packages(packages=None, present=True, pkg_path=None, state=None, host=None):
    '''
    Install/remove/update pkg packages. This will use ``pkg ...`` where available
    (FreeBSD) and the ``pkg_*`` variants elsewhere.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + pkg_path: the PKG_PATH environment variable to set

    pkg_path:
        By default this is autogenerated as follows (tested/working for OpenBSD):
        ``http://ftp.<OS>.org/pub/<OS>/<VERSION>/packages/<ARCH>/``. Note that OpenBSD's
        official mirrors only hold the latest two versions packages.

        NetBSD/FreeBSD helpfully use their own directory structures, so the default won't
        work.

    Example:

    .. code:: python

        pkg.packages(
            name='Install Vim and Vim Addon Manager',
            packages=['vim-addon-manager', 'vim'],
        )

    '''

    if present is True:
        if not pkg_path and not host.get_fact(File, path='/etc/installurl'):
            host_os = host.get_fact(Os) or ''
            pkg_path = 'http://ftp.{http}.org/pub/{os}/{version}/packages/{arch}/'.format(
                http=host_os.lower(),
                os=host_os,
                version=host.get_fact(OsVersion),
                arch=host.get_fact(Arch),
            )

    # FreeBSD used "pkg ..." and OpenBSD uses "pkg_[add|delete]"
    is_pkg = host.get_fact(Which, command='pkg')
    install_command = 'pkg install -y' if is_pkg else 'pkg_add'
    uninstall_command = 'pkg delete -y' if is_pkg else 'pkg_delete'

    if pkg_path:
        install_command = 'PKG_PATH={0} {1}'.format(pkg_path, install_command)

    yield ensure_packages(
        host, packages, host.get_fact(PkgPackages), present,
        install_command=install_command,
        uninstall_command=uninstall_command,
    )



from __future__ import unicode_literals

import six

from six import StringIO
from six.moves import shlex_quote
from six.moves.urllib.parse import urlparse

from pyinfra.facts.files import File
from pyinfra.facts.rpm import RpmPackage


def _has_package(package, packages, expand_package_fact=None, match_any=False):
    def in_packages(pkg):
        if isinstance(pkg, list):
            return pkg[0] in packages and pkg[1] in packages[pkg[0]]
        return pkg in packages

    packages_to_check = [package]
    if expand_package_fact:
        packages_to_check = expand_package_fact(package) or packages_to_check

    checks = (in_packages(pkg) for pkg in packages_to_check)
    if match_any:
        return any(checks)
    return all(checks)


def ensure_packages(
    host, packages, current_packages, present,
    install_command, uninstall_command,
    latest=False, upgrade_command=None,
    version_join=None, lower=True,
    expand_package_fact=None,
):
    '''
    Handles this common scenario:

    + We have a list of packages(/versions) to ensure
    + We have a map of existing package -> versions
    + We have the common command bits (install, uninstall, version "joiner")
    + Outputs commands to ensure our desired packages/versions
    + Optionally upgrades packages w/o specified version when present

    Args:
        packages (list): list of packages or package/versions
        current_packages (fact): fact returning dict of package names -> version
        present (bool): whether packages should exist or not
        install_command (str): command to prefix to list of packages to install
        uninstall_command (str): as above for uninstalling packages
        latest (bool): whether to upgrade installed packages when present
        upgrade_command (str): as above for upgrading
        version_join (str): the package manager specific "joiner", ie ``=`` for \
            ``<apt_pkg>=<version>``
        lower (bool): whether to lowercase package names
    '''

    if packages is None:
        return

    # Accept a single package as string
    if isinstance(packages, six.string_types):
        packages = [packages]

    # Lowercase packaging?
    if lower:
        packages = [
            package.lower()
            for package in packages
        ]

    # Version support?
    if version_join:
        # Split where versions present
        packages = [
            package.rsplit(version_join, 1)
            for package in packages
        ]

        # Covert to either string or list
        packages = [
            package[0] if len(package) == 1
            else package
            for package in packages
        ]

    # Diff the ensured packages against the remote state/fact
    diff_packages = []

    # Packages to upgrade? (install only)
    upgrade_packages = []

    # Installing?
    if present is True:
        for package in packages:
            # String version, just check if not existing
            if not _has_package(package, current_packages, expand_package_fact):
                diff_packages.append(package)

            else:
                # Present packages w/o version specified - for upgrade if latest
                if isinstance(package, six.string_types):
                    upgrade_packages.append(package)

                if not latest:
                    pkg_key = package[0] if isinstance(package, list) else package
                    if pkg_key in current_packages:
                        host.noop('package {0} is installed ({1})'.format(
                            package, ', '.join(current_packages[pkg_key]),
                        ))
                    else:
                        host.noop('package {0} is installed'.format(package))

    # Uninstalling?
    else:
        for package in packages:
            # String version, just check if existing
            if _has_package(package, current_packages, expand_package_fact, match_any=True):
                diff_packages.append(package)

            else:
                host.noop('package {0} is not installed'.format(package))

    # Convert packages back to string(/version)
    diff_packages = [
        version_join.join(package)
        if isinstance(package, list)
        else package
        for package in diff_packages
    ]

    if diff_packages:
        command = install_command if present else uninstall_command

        yield '{0} {1}'.format(
            command,
            ' '.join([shlex_quote(pkg) for pkg in diff_packages]),
        )

        for package in diff_packages:  # add/remove from current packages
            if present:
                version = 'unknown'
                if version_join:
                    bits = package.rsplit(version_join, 1)
                    package = bits[0]
                    if len(bits) == 2:
                        version = bits[1]
                current_packages[package] = [version]
            else:
                current_packages.pop(package, None)

    if latest and upgrade_command and upgrade_packages:
        yield '{0} {1}'.format(
            upgrade_command,
            ' '.join([shlex_quote(pkg) for pkg in upgrade_packages]),
        )


def ensure_rpm(state, host, files, source, present, package_manager_command):
    original_source = source

    # If source is a url
    if urlparse(source).scheme:
        # Generate a temp filename (with .rpm extension to please yum)
        temp_filename = '{0}.rpm'.format(state.get_temp_filename(source))

        # Ensure it's downloaded
        yield files.download(source, temp_filename, state=state, host=host)

        # Override the source with the downloaded file
        source = temp_filename

    # Check for file .rpm information
    info = host.get_fact(RpmPackage, name=source)
    exists = False

    # We have info!
    if info:
        current_package = host.get_fact(RpmPackage, name=info['name'])
        if current_package and current_package['version'] == info['version']:
            exists = True

    # Package does not exist and we want?
    if present and not exists:
        # If we had info, always install
        if info:
            yield 'rpm -i {0}'.format(source)

        # This happens if we download the package mid-deploy, so we have no info
        # but also don't know if it's installed. So check at runtime, otherwise
        # the install will fail.
        else:
            yield 'rpm -q `rpm -qp {0}` 2> /dev/null || rpm -i {0}'.format(source)

    # Package exists but we don't want?
    elif exists and not present:
        yield '{0} remove -y {1}'.format(package_manager_command, info['name'])

    else:
        host.noop('rpm {0} is {1}'.format(
            original_source,
            'installed' if present else 'not installed',
        ))


def ensure_yum_repo(
    state, host, files,
    name_or_url, baseurl, present, description, enabled, gpgcheck, gpgkey,
    repo_directory='/etc/yum.repos.d/',
    type_=None,
):
    url = None
    url_parts = urlparse(name_or_url)
    if url_parts.scheme:
        url = name_or_url
        name_or_url = url_parts.path.split('/')[-1]
        if name_or_url.endswith('.repo'):
            name_or_url = name_or_url[:-5]

    filename = '{0}{1}.repo'.format(repo_directory, name_or_url)

    # If we don't want the repo, just remove any existing file
    if not present:
        yield files.file(filename, present=False, state=state, host=host)
        return

    # If we're a URL, download the repo if it doesn't exist
    if url:
        if not host.get_fact(File, path=filename):
            yield files.download(url, filename, state=state, host=host)
        return

    # Description defaults to name
    description = description or name_or_url

    # Build the repo file from string
    repo_lines = [
        '[{0}]'.format(name_or_url),
        'name={0}'.format(description),
        'baseurl={0}'.format(baseurl),
        'enabled={0}'.format(1 if enabled else 0),
        'gpgcheck={0}'.format(1 if gpgcheck else 0),
    ]

    if type_:
        repo_lines.append('type={0}'.format(type_))

    if gpgkey:
        repo_lines.append('gpgkey={0}'.format(gpgkey))

    repo_lines.append('')
    repo = '\n'.join(repo_lines)
    repo = StringIO(repo)

    # Ensure this is the file on the server
    yield files.put(repo, filename, state=state, host=host)



'''
Manage launchd services.
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.launchd import LaunchdStatus

from .util.service import handle_service_control


@operation
def service(
    service,
    running=True, restarted=False, command=None,
    state=None, host=None,
):
    '''
    Manage the state of systemd managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + command: custom command to pass like: ``launchctl <command> <service>``
    + enabled: whether this service should be enabled/disabled on boot
    '''

    was_running = host.get_fact(LaunchdStatus).get(service, None)

    yield handle_service_control(
        host,
        service, LaunchdStatus,
        'launchctl {1} {0}',
        # No support for restart/reload/command
        running, None, None, None,
    )

    # No restart command, so just stop/start
    if restarted and was_running:
        yield 'launchctl stop {0}'.format(service)
        yield 'launchctl start {0}'.format(service)



'''
Manage Ruby gem packages. (see https://rubygems.org/ )
'''

from pyinfra.api import operation
from pyinfra.facts.gem import GemPackages

from .util.packaging import ensure_packages


@operation
def packages(packages=None, present=True, latest=False, state=None, host=None):
    '''
    Add/remove/update gem packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.

    Example:

    .. code:: python

        # Note: Assumes that 'gem' is installed.
        gem.packages(
            name='Install rspec',
            packages=['rspec'],
        )
    '''

    yield ensure_packages(
        host, packages, host.get_fact(GemPackages), present,
        install_command='gem install',
        uninstall_command='gem uninstall',
        upgrade_command='gem update',
        version_join=':',
        latest=latest,
    )



from __future__ import print_function

from unittest import TestCase

try:
    from pathlib import PurePosixPath
    HAS_PATHLIB = True
except ImportError:
    HAS_PATHLIB = False

import pytest

from pyinfra.api import (
    FileDownloadCommand,
    FileUploadCommand,
    FunctionCommand,
    MaskString,
    QuoteString,
    RsyncCommand,
    StringCommand,
)
from pyinfra.api.command import PyinfraCommand


class TestBaseCommand(TestCase):
    def test_base_command_no_execute(self):
        cmd = PyinfraCommand()
        with self.assertRaises(NotImplementedError):
            cmd.execute(None, None, None)


class TestStringCommand(TestCase):
    def test_normal(self):
        cmd = StringCommand('hello', 'world')
        assert str(cmd) == cmd.get_raw_value() == 'hello world'

    def test_masked(self):
        cmd = StringCommand(MaskString('adsfg'))
        assert cmd.get_raw_value() == 'adsfg'
        assert str(cmd) == '***'

    def test_mixed_masked(self):
        cmd = StringCommand('some', 'stuff', MaskString('mask me'), 'other', 'stuff')
        assert cmd.get_raw_value() == 'some stuff mask me other stuff'
        assert str(cmd) == 'some stuff *** other stuff'

    def test_nested(self):
        nested_cmd = StringCommand('some', 'stuff')
        cmd = StringCommand('hello', nested_cmd, 'world')
        assert str(cmd) == cmd.get_raw_value() == 'hello some stuff world'

    def test_quote(self):
        quoted_str = QuoteString('quote me!')
        cmd = StringCommand('hello', quoted_str)
        assert str(cmd) == cmd.get_raw_value() == "hello 'quote me!'"

    def test_eq(self):
        assert StringCommand('hello world') == StringCommand('hello world')
        with self.assertRaises(AssertionError):
            assert StringCommand('a') == StringCommand('b')

    def test_separator(self):
        cmd = StringCommand('hello world', 'yes', _separator='')
        assert str(cmd) == 'hello worldyes'

    def test_int(self):
        cmd = StringCommand('hello', 'world', 4420)
        assert str(cmd) == 'hello world 4420'

    def test_list(self):
        cmd = StringCommand('hello', 'world', ['a', 'b', 'c'])
        assert str(cmd) == "hello world ['a', 'b', 'c']"

    @pytest.mark.skipif(
        not HAS_PATHLIB,
        reason='Requires Python 3.4+ (pathlib module)',
    )
    def test_path(self):
        cmd = StringCommand('hello', 'world', PurePosixPath('/path/to/somewhere'))
        assert str(cmd) == 'hello world /path/to/somewhere'


class TestFileCommands(TestCase):
    def test_file_upload_command_repr(self):
        cmd = FileUploadCommand('src', 'dest')
        assert repr(cmd) == 'FileUploadCommand(src, dest)'

    def test_file_download_command_repr(self):
        cmd = FileDownloadCommand('src', 'dest')
        assert repr(cmd) == 'FileDownloadCommand(src, dest)'

    def test_rsync_command_repr(self):
        cmd = RsyncCommand('src', 'dest', ['-a'])
        assert repr(cmd) == "RsyncCommand(src, dest, ['-a'])"


class TestFunctionCommand(TestCase):
    def test_function_command_repr(self):
        def some_function():
            pass

        cmd = FunctionCommand(some_function, (), {})
        assert repr(cmd) == 'FunctionCommand(some_function, (), {})'



from __future__ import print_function, unicode_literals

import base64
import ntpath

import click

from pyinfra import logger
from pyinfra.api import Config
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.api.util import get_file_io, memoize, sha1_hash

from .pyinfrawinrmsession import PyinfraWinrmSession
from .util import make_win_command


def _raise_connect_error(host, message, data):
    message = '{0} ({1})'.format(message, data)
    raise ConnectError(message)


@memoize
def show_warning():
    logger.warning('The @winrm connector is alpha!')


def _make_winrm_kwargs(state, host):
    kwargs = {
    }

    for key, value in (
        ('username', host.data.winrm_user),
        ('password', host.data.winrm_password),
        ('winrm_port', int(host.data.winrm_port or 0)),
        ('winrm_transport', host.data.winrm_transport or 'plaintext'),
        ('winrm_read_timeout_sec', host.data.winrm_read_timeout_sec or 30),
        ('winrm_operation_timeout_sec', host.data.winrm_operation_timeout_sec or 20),
    ):
        if value:
            kwargs[key] = value

    # FUTURE: add more auth
    # pywinrm supports: basic, certificate, ntlm, kerberos, plaintext, ssl, credssp
    # see https://github.com/diyan/pywinrm/blob/master/winrm/__init__.py#L12

    return kwargs


def make_names_data(hostname):

    show_warning()

    yield '@winrm/{0}'.format(hostname), {'winrm_hostname': hostname}, []


def connect(state, host):
    '''
    Connect to a single host. Returns the winrm Session if successful.
    '''

    kwargs = _make_winrm_kwargs(state, host)
    logger.debug('Connecting to: %s (%s)', host.name, kwargs)

    # Hostname can be provided via winrm config (alias), data, or the hosts name
    hostname = kwargs.pop(
        'hostname',
        host.data.winrm_hostname or host.name,
    )

    try:
        # Create new session
        host_and_port = '{}:{}'.format(hostname, host.data.winrm_port)
        logger.debug('host_and_port: %s', host_and_port)

        session = PyinfraWinrmSession(
            host_and_port,
            auth=(
                kwargs['username'],
                kwargs['password'],
            ),
            transport=kwargs['winrm_transport'],
            read_timeout_sec=kwargs['winrm_read_timeout_sec'],
            operation_timeout_sec=kwargs['winrm_operation_timeout_sec'],
        )

        return session

    # TODO: add exceptions here
    except Exception as e:
        auth_kwargs = {}

        for key, value in kwargs.items():
            if key in ('username', 'password'):
                auth_kwargs[key] = value

        auth_args = ', '.join(
            '{0}={1}'.format(key, value)
            for key, value in auth_kwargs.items()
        )
        logger.debug('%s', e)
        _raise_connect_error(host, 'Authentication error', auth_args)


def run_shell_command(
    state, host, command,
    env=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    shell_executable=Config.SHELL,
    **ignored_command_kwargs
):
    '''
    Execute a command on the specified host.

    Args:
        state (``pyinfra.api.State`` obj): state object for this command
        hostname (string): hostname of the target
        command (string): actual command to execute
        success_exit_codes (list): all values in the list that will return success
        print_output (boolean): print the output
        print_intput (boolean): print the input
        return_combined_output (boolean): combine the stdout and stderr lists
        shell_executable (string): shell to use - 'cmd'=cmd, 'ps'=powershell(default)
        env (dict): environment variables to set

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    '''

    command = make_win_command(command)

    logger.debug('Running command on %s: %s', host.name, command)

    if print_input:
        click.echo('{0}>>> {1}'.format(host.print_prefix, command), err=True)

    # get rid of leading/trailing quote
    tmp_command = command.strip("'")

    if print_output:
        click.echo(
            '{0}>>> {1}'.format(host.print_prefix, command),
            err=True,
        )

    if not shell_executable:
        shell_executable = 'ps'
    logger.debug('shell_executable:%s', shell_executable)

    # we use our own subclassed session that allows for env setting from open_shell.
    if shell_executable in ['cmd']:
        response = host.connection.run_cmd(tmp_command, env=env)
    else:
        response = host.connection.run_ps(tmp_command, env=env)

    return_code = response.status_code
    logger.debug('response:%s', response)

    std_out_str = response.std_out.decode('utf-8')
    std_err_str = response.std_err.decode('utf-8')

    # split on '\r\n' (windows newlines)
    std_out = std_out_str.split('\r\n')
    std_err = std_err_str.split('\r\n')

    logger.debug('std_out:%s', std_out)
    logger.debug('std_err:%s', std_err)

    if print_output:
        click.echo(
            '{0}>>> {1}'.format(host.print_prefix, '\n'.join(std_out)),
            err=True,
        )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    logger.debug('Command exit status: %s', status)

    if return_combined_output:
        std_out = [('stdout', line) for line in std_out]
        std_err = [('stderr', line) for line in std_err]
        return status, std_out + std_err

    return status, std_out, std_err


def get_file(
    state, host, remote_filename, filename_or_io,
    **command_kwargs
):
    raise PyinfraError('Not implemented')


def _put_file(state, host, filename_or_io, remote_location, chunk_size=2048):
    # this should work fine on smallish files, but there will be perf issues
    # on larger files both due to the full read, the base64 encoding, and
    # the latency when sending chunks
    with get_file_io(filename_or_io) as file_io:
        data = file_io.read()
        for i in range(0, len(data), chunk_size):
            chunk = data[i:i + chunk_size]
            ps = (
                '$data = [System.Convert]::FromBase64String("{0}"); '
                '{1} -Value $data -Encoding byte -Path "{2}"'
            ).format(
                base64.b64encode(chunk).decode('utf-8'),
                'Set-Content' if i == 0 else 'Add-Content',
                remote_location)
            status, _stdout, stderr = run_shell_command(state, host, ps)
            if status is False:
                logger.error('File upload error: {0}'.format('\n'.join(stderr)))
                return False

    return True


def put_file(
    state, host, filename_or_io, remote_filename,
    print_output=False, print_input=False,
    **command_kwargs
):
    '''
    Upload file by chunking and sending base64 encoded via winrm
    '''

    # Always use temp file here in case of failure
    temp_file = ntpath.join(
        host.fact.windows_temp_dir(),
        'pyinfra-{0}'.format(sha1_hash(remote_filename)),
    )

    if not _put_file(state, host, filename_or_io, temp_file):
        return False

    # Execute run_shell_command w/sudo and/or su_user
    command = 'Move-Item -Path {0} -Destination {1} -Force'.format(temp_file, remote_filename)
    status, _, stderr = run_shell_command(
        state, host, command,
        print_output=print_output,
        print_input=print_input,
        **command_kwargs
    )

    if status is False:
        logger.error('File upload error: {0}'.format('\n'.join(stderr)))
        return False

    if print_output:
        click.echo(
            '{0}file uploaded: {1}'.format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


EXECUTION_CONNECTOR = True



from __future__ import print_function, unicode_literals

from distutils.spawn import find_executable
from getpass import getpass
from os import path
from socket import (
    error as socket_error,
    gaierror,
)

import click
import six

from paramiko import (
    AuthenticationException,
    DSSKey,
    ECDSAKey,
    Ed25519Key,
    MissingHostKeyPolicy,
    PasswordRequiredException,
    RSAKey,
    SFTPClient,
    SSHException,
)

import pyinfra

from pyinfra import logger
from pyinfra.api.command import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.api.util import get_file_io, memoize

from .sshuserclient import SSHClient
from .util import (
    get_sudo_password,
    make_unix_command,
    read_buffers_into_queue,
    run_local_process,
    split_combined_output,
    write_stdin,
)

EXECUTION_CONNECTOR = True


def make_names_data(hostname):
    yield '@ssh/{0}'.format(hostname), {'ssh_hostname': hostname}, []


def _raise_connect_error(host, message, data):
    message = '{0} ({1})'.format(message, data)
    raise ConnectError(message)


def _load_private_key_file(filename, key_filename, key_password):
    exception = PyinfraError('Invalid key: {0}'.format(filename))

    for key_cls in (RSAKey, DSSKey, ECDSAKey, Ed25519Key):
        try:
            return key_cls.from_private_key_file(
                filename=filename,
            )

        except PasswordRequiredException:
            if not key_password:
                # If password is not provided, but we're in CLI mode, ask for it. I'm not a
                # huge fan of having CLI specific code in here, but it doesn't really fit
                # anywhere else without duplicating lots of key related code into cli.py.
                if pyinfra.is_cli:
                    key_password = getpass(
                        'Enter password for private key: {0}: '.format(
                            key_filename,
                        ),
                    )

                # API mode and no password? We can't continue!
                else:
                    raise PyinfraError(
                        'Private key file ({0}) is encrypted, set ssh_key_password to '
                        'use this key'.format(key_filename),
                    )

            try:
                return key_cls.from_private_key_file(
                    filename=filename,
                    password=key_password,
                )
            except SSHException as e:  # key does not match key_cls type
                exception = e
        except SSHException as e:  # key does not match key_cls type
            exception = e
    raise exception


def _get_private_key(state, key_filename, key_password):
    if key_filename in state.private_keys:
        return state.private_keys[key_filename]

    ssh_key_filenames = [
        # Global from executed directory
        path.expanduser(key_filename),
    ]

    # Relative to the deploy
    if state.deploy_dir:
        ssh_key_filenames.append(
            path.join(state.deploy_dir, key_filename),
        )

    key = False
    key_file_exists = False

    for filename in ssh_key_filenames:
        if not path.isfile(filename):
            continue

        key_file_exists = True

        try:
            key = _load_private_key_file(filename, key_filename, key_password)
            break
        except SSHException:
            pass

    # No break, so no key found
    if not key:
        if not key_file_exists:
            raise PyinfraError('No such private key file: {0}'.format(key_filename))

        # TODO: upgrade min paramiko version to 2.7 and remove this (pyinfra v2)
        extra_info = ''
        from pkg_resources import get_distribution, parse_version
        if get_distribution('paramiko').parsed_version < parse_version('2.7'):
            extra_info = (
                '\n    Paramiko versions under 2.7 do not support the latest OpenSSH key formats,'
                ' upgrading may fix this error.'
                '\n    For more information, see this issue: '
                'https://github.com/Fizzadar/pyinfra/issues/548'
            )
        raise PyinfraError('Invalid private key file: {0}{1}'.format(key_filename, extra_info))

    # Load any certificate, names from OpenSSH:
    # https://github.com/openssh/openssh-portable/blob/049297de975b92adcc2db77e3fb7046c0e3c695d/ssh-keygen.c#L2453  # noqa: E501
    for certificate_filename in (
        '{0}-cert.pub'.format(key_filename),
        '{0}.pub'.format(key_filename),
    ):
        if path.isfile(certificate_filename):
            key.load_certificate(certificate_filename)

    state.private_keys[key_filename] = key
    return key


def _make_paramiko_kwargs(state, host):
    kwargs = {
        'allow_agent': False,
        'look_for_keys': False,
        'hostname': host.data.ssh_hostname or host.name,
    }

    for key, value in (
        ('username', host.data.ssh_user),
        ('port', int(host.data.ssh_port or 0)),
        ('timeout', state.config.CONNECT_TIMEOUT),
    ):
        if value:
            kwargs[key] = value

    # Password auth (boo!)
    if host.data.ssh_password:
        kwargs['password'] = host.data.ssh_password

    # Key auth!
    elif host.data.ssh_key:
        kwargs['pkey'] = _get_private_key(
            state,
            key_filename=host.data.ssh_key,
            key_password=host.data.ssh_key_password,
        )

    # No key or password, so let's have paramiko look for SSH agents and user keys
    else:
        kwargs['allow_agent'] = True
        kwargs['look_for_keys'] = True

    return kwargs


def connect(state, host):
    '''
    Connect to a single host. Returns the SSH client if succesful. Stateless by
    design so can be run in parallel.
    '''

    kwargs = _make_paramiko_kwargs(state, host)
    logger.debug('Connecting to: {0} ({1})'.format(host.name, kwargs))

    hostname = kwargs.pop('hostname')

    try:
        # Create new client & connect to the host
        client = SSHClient()
        client.set_missing_host_key_policy(MissingHostKeyPolicy())
        client.connect(hostname, **kwargs)
        return client

    except AuthenticationException:
        auth_kwargs = {}

        for key, value in kwargs.items():
            if key in ('username', 'password'):
                auth_kwargs[key] = value
                continue

            if key == 'pkey' and value:
                auth_kwargs['key'] = host.data.ssh_key

        auth_args = ', '.join(
            '{0}={1}'.format(key, value)
            for key, value in auth_kwargs.items()
        )

        _raise_connect_error(host, 'Authentication error', auth_args)

    except SSHException as e:
        _raise_connect_error(host, 'SSH error', e)

    except gaierror:
        _raise_connect_error(host, 'Could not resolve hostname', hostname)

    except socket_error as e:
        _raise_connect_error(host, 'Could not connect', e)

    except EOFError as e:
        _raise_connect_error(host, 'EOF error', e)


def run_shell_command(
    state, host, command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    use_sudo_password=False,
    **command_kwargs
):
    '''
    Execute a command on the specified host.

    Args:
        state (``pyinfra.api.State`` obj): state object for this command
        hostname (string): hostname of the target
        command (string): actual command to execute
        sudo (boolean): whether to wrap the command with sudo
        sudo_user (string): user to sudo to
        get_pty (boolean): whether to get a PTY before executing the command
        env (dict): environment variables to set
        timeout (int): timeout for this command to complete before erroring

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    '''

    if use_sudo_password:
        command_kwargs['use_sudo_password'] = get_sudo_password(
            state, host, use_sudo_password,
            run_shell_command=run_shell_command,
            put_file=put_file,
        )

    command = make_unix_command(command, state=state, **command_kwargs)
    actual_command = command.get_raw_value()

    logger.debug('Running command on {0}: (pty={1}) {2}'.format(
        host.name, get_pty, command,
    ))

    if print_input:
        click.echo('{0}>>> {1}'.format(host.print_prefix, command), err=True)

    # Run it! Get stdout, stderr & the underlying channel
    stdin_buffer, stdout_buffer, stderr_buffer = host.connection.exec_command(
        actual_command,
        get_pty=get_pty,
    )

    if stdin:
        write_stdin(stdin, stdin_buffer)

    combined_output = read_buffers_into_queue(
        stdout_buffer,
        stderr_buffer,
        timeout=timeout,
        print_output=print_output,
        print_prefix=host.print_prefix,
    )

    logger.debug('Waiting for exit status...')
    exit_status = stdout_buffer.channel.recv_exit_status()
    logger.debug('Command exit status: {0}'.format(exit_status))

    if success_exit_codes:
        status = exit_status in success_exit_codes
    else:
        status = exit_status == 0

    if return_combined_output:
        return status, combined_output

    stdout, stderr = split_combined_output(combined_output)
    return status, stdout, stderr


@memoize
def _get_sftp_connection(host):
    transport = host.connection.get_transport()

    try:
        return SFTPClient.from_transport(transport)
    except SSHException as e:
        six.raise_from(ConnectError((
            'Unable to establish SFTP connection. Check that the SFTP subsystem '
            'for the SSH service at {0} is enabled.'
        ).format(host)), e)


def _get_file(host, remote_filename, filename_or_io):
    with get_file_io(filename_or_io, 'wb') as file_io:
        sftp = _get_sftp_connection(host)
        sftp.getfo(remote_filename, file_io)


def get_file(
    state, host, remote_filename, filename_or_io,
    sudo=False, sudo_user=None, su_user=None,
    print_output=False, print_input=False,
    **command_kwargs
):
    '''
    Download a file from the remote host using SFTP. Supports download files
    with sudo by copying to a temporary directory with read permissions,
    downloading and then removing the copy.
    '''

    if sudo or su_user:
        # Get temp file location
        temp_file = state.get_temp_filename(remote_filename)

        # Copy the file to the tempfile location and add read permissions
        command = 'cp {0} {1} && chmod +r {0}'.format(remote_filename, temp_file)

        copy_status, _, stderr = run_shell_command(
            state, host, command,
            sudo=sudo, sudo_user=sudo_user, su_user=su_user,
            print_output=print_output,
            print_input=print_input,
            **command_kwargs
        )

        if copy_status is False:
            logger.error('File download copy temp error: {0}'.format('\n'.join(stderr)))
            return False

        try:
            _get_file(host, temp_file, filename_or_io)

        # Ensure that, even if we encounter an error, we (attempt to) remove the
        # temporary copy of the file.
        finally:
            remove_status, _, stderr = run_shell_command(
                state, host, 'rm -f {0}'.format(temp_file),
                sudo=sudo, sudo_user=sudo_user, su_user=su_user,
                print_output=print_output,
                print_input=print_input,
                **command_kwargs
            )

        if remove_status is False:
            logger.error('File download remove temp error: {0}'.format('\n'.join(stderr)))
            return False

    else:
        _get_file(host, remote_filename, filename_or_io)

    if print_output:
        click.echo(
            '{0}file downloaded: {1}'.format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def _put_file(host, filename_or_io, remote_location):
    with get_file_io(filename_or_io) as file_io:
        sftp = _get_sftp_connection(host)
        sftp.putfo(file_io, remote_location)


def put_file(
    state, host, filename_or_io, remote_filename,
    sudo=False, sudo_user=None, su_user=None,
    print_output=False, print_input=False,
    **command_kwargs
):
    '''
    Upload file-ios to the specified host using SFTP. Supports uploading files
    with sudo by uploading to a temporary directory then moving & chowning.
    '''

    # sudo/su are a little more complicated, as you can only sftp with the SSH
    # user connected, so upload to tmp and copy/chown w/sudo and/or su_user
    if sudo or su_user:
        # Get temp file location
        temp_file = state.get_temp_filename(remote_filename)
        _put_file(host, filename_or_io, temp_file)

        # Execute run_shell_command w/sudo and/or su_user
        command = StringCommand('mv', temp_file, QuoteString(remote_filename))

        # Move it to the su_user if present
        if su_user:
            command = StringCommand(command, '&&', 'chown', su_user, QuoteString(remote_filename))

        # Otherwise any sudo_user
        elif sudo_user:
            command = StringCommand(command, '&&', 'chown', sudo_user, QuoteString(remote_filename))

        status, _, stderr = run_shell_command(
            state, host, command,
            sudo=sudo, sudo_user=sudo_user, su_user=su_user,
            print_output=print_output,
            print_input=print_input,
            **command_kwargs
        )

        if status is False:
            logger.error('File upload error: {0}'.format('\n'.join(stderr)))
            return False

    # No sudo and no su_user, so just upload it!
    else:
        _put_file(host, filename_or_io, remote_filename)

    if print_output:
        click.echo(
            '{0}file uploaded: {1}'.format(host.print_prefix, remote_filename),
            err=True,
        )

    return True


def check_can_rsync(host):
    if host.data.ssh_key_password:
        raise NotImplementedError('Rsync does not currently work with SSH keys needing passwords.')

    if host.data.ssh_password:
        raise NotImplementedError('Rsync does not currently work with SSH passwords.')

    if not find_executable('rsync'):
        raise NotImplementedError('The `rsync` binary is not available on this system.')


def rsync(
    state, host, src, dest, flags,
    print_output=False, print_input=False,
    sudo=False,
    sudo_user=None,
    **ignored_kwargs
):
    hostname = host.data.ssh_hostname or host.name
    user = ''
    if host.data.ssh_user:
        user = '{0}@'.format(host.data.ssh_user)

    ssh_flags = []

    port = host.data.ssh_port
    if port:
        ssh_flags.append('-p {0}'.format(port))

    ssh_key = host.data.ssh_key
    if ssh_key:
        ssh_flags.append('-i {0}'.format(ssh_key))

    remote_rsync_command = 'rsync'
    if sudo:
        remote_rsync_command = 'sudo rsync'
        if sudo_user:
            remote_rsync_command = 'sudo -u {0} rsync'.format(sudo_user)

    rsync_command = (
        'rsync {rsync_flags} '
        "--rsh 'ssh -o BatchMode=yes -o StrictHostKeyChecking=no {ssh_flags}' "
        "--rsync-path '{remote_rsync_command}' "
        '{src} {user}{hostname}:{dest}'
    ).format(
        rsync_flags=' '.join(flags),
        ssh_flags=' '.join(ssh_flags),
        remote_rsync_command=remote_rsync_command,
        user=user, hostname=hostname,
        src=src, dest=dest,
    )

    if print_input:
        click.echo('{0}>>> {1}'.format(host.print_prefix, rsync_command), err=True)

    return_code, combined_output = run_local_process(
        rsync_command,
        print_output=print_output,
        print_prefix=host.print_prefix,
    )

    status = return_code == 0

    if not status:
        _, stderr = split_combined_output(combined_output)
        raise IOError('\n'.join(stderr))

    return True



def make_cat_files_command(*filenames):
    commands = []

    for filename in filenames:
        if '*' in filename:
            # There's no way to test against a glob expression, so accept anything here
            commands.append('cat {0} || true'.format(filename))
        else:
            commands.append('! test -f {0} || cat {0}'.format(filename))

    if len(commands) > 1:  # if we have multiple, wrap them
        commands = ['({0})'.format(command) for command in commands]

    return ' && '.join(commands)



from pyinfra.api import FactBase

from .util import make_cat_files_command
from .util.packaging import parse_yum_repositories


class YumRepositories(FactBase):
    '''
    Returns a list of installed yum repositories:

    .. code:: python

        [
            {
                'name': 'CentOS-$releasever - Apps',
                'baseurl': 'http://mirror.centos.org/$contentdir/$releasever/Apps/$basearch/os/',
                'gpgcheck': '1',
                'enabled': '1',
                'gpgkey': 'file:///etc/pki/rpm-gpg/RPM-GPG-KEY-centosofficial',
            },
        ]
    '''

    command = make_cat_files_command(
        '/etc/yum.conf',
        '/etc/yum.repos.d/*.repo',
    )
    requires_command = 'yum'

    default = list

    def process(self, output):
        return parse_yum_repositories(output)



'''
Execute commands and up/download files *from* the remote host.

Eg: ``pyinfra -> inventory-host.net <-> another-host.net``
'''

from six.moves import shlex_quote

from pyinfra import logger
from pyinfra.api import operation, OperationError
from pyinfra.facts.files import File, FindInFile

from . import files


@operation
def keyscan(hostname, force=False, port=22, state=None, host=None):
    '''
    Check/add hosts to the ``~/.ssh/known_hosts`` file.

    + hostname: hostname that should have a key in ``known_hosts``
    + force: if the key already exists, remove and rescan

    Example:

    .. code:: python

        ssh.keyscan(
            name='Set add server two to known_hosts on one',
            hostname='two.example.com',
        )
    '''

    yield files.directory(
        '~/.ssh',
        mode=700,
        state=state,
        host=host,
    )

    hostname_present = host.get_fact(
        FindInFile,
        path='~/.ssh/known_hosts',
        pattern=hostname,
    )

    keyscan_command = 'ssh-keyscan -p {0} {1} >> ~/.ssh/known_hosts'.format(
        port, hostname,
    )

    if not hostname_present:
        yield keyscan_command

    elif force:
        yield 'ssh-keygen -R {0}'.format(hostname)
        yield keyscan_command

    else:
        host.noop('host key for {0} already exists'.format(hostname))


def _user_or_ssh_user(user, ssh_user):
    if ssh_user:
        logger.warning((
            'The `ssh_user` argument is deprecated in `ssh.*` operations, '
            'please use the `user` argument.'
        ))
    return user or ssh_user


@operation
def command(hostname, command, user=None, port=22, ssh_user=None, state=None, host=None):
    '''
    Execute commands on other servers over SSH.

    + hostname: the hostname to connect to
    + command: the command to execute
    + user: connect with this user
    + port: connect to this port

    Example:

    .. code:: python

        ssh.command(
            name='Create file by running echo from host one to host two',
            hostname='two.example.com',
            command='echo "one was here" > /tmp/one.txt',
            user='vagrant',
        )
    '''

    # TODO: remove this (ssh_user is a legacy arg)
    user = _user_or_ssh_user(user, ssh_user)

    command = shlex_quote(command)

    connection_target = hostname
    if user:
        connection_target = '@'.join((user, hostname))

    yield 'ssh -p {0} {1} {2}'.format(port, connection_target, command)


@operation
def upload(
    hostname, filename,
    remote_filename=None,
    port=22,
    user=None,
    use_remote_sudo=False,
    ssh_keyscan=False,
    ssh_user=None,
    state=None, host=None,
):
    '''
    Upload files to other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to upload
    + remote_filename: where to upload the file to (defaults to ``filename``)
    + port: connect to this port
    + user: connect with this user
    + use_remote_sudo: upload to a temporary location and move using sudo
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    '''

    remote_filename = remote_filename or filename

    # TODO: remove this (ssh_user is a legacy arg)
    user = _user_or_ssh_user(user, ssh_user)

    # Figure out where we're connecting (host or user@host)
    connection_target = hostname
    if user:
        connection_target = '@'.join((user, hostname))

    if ssh_keyscan:
        yield keyscan(hostname, state=state, host=host)

    # If we're not using sudo on the remote side, just scp the file over
    if not use_remote_sudo:
        yield 'scp -P {0} {1} {2}:{3}'.format(
            port, filename, connection_target, remote_filename,
        )

    else:
        # Otherwise - we need a temporary location for the file
        temp_remote_filename = state.get_temp_filename()

        # scp it to the temporary location
        upload_cmd = 'scp -P {0} {1} {2}:{3}'.format(
            port, filename, connection_target, temp_remote_filename,
        )

        yield upload_cmd

        # And sudo sudo to move it
        yield command(
            hostname=hostname,
            command='sudo mv {0} {1}'.format(temp_remote_filename, remote_filename),
            port=port,
            user=user,
            state=state,
            host=host,
        )


@operation
def download(
    hostname, filename,
    local_filename=None,
    force=False,
    port=22,
    user=None,
    ssh_keyscan=False,
    ssh_user=None,
    state=None, host=None,
):
    '''
    Download files from other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to download
    + local_filename: where to download the file to (defaults to ``filename``)
    + force: always download the file, even if present locally
    + port: connect to this port
    + user: connect with this user
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    '''

    local_filename = local_filename or filename

    # TODO: remove this (ssh_user is a legacy arg)
    user = _user_or_ssh_user(user, ssh_user)

    # Get local file info
    local_file_info = host.get_fact(File, path=local_filename)

    # Local file exists but isn't a file?
    if local_file_info is False:
        raise OperationError(
            'Local destination {0} already exists and is not a file'.format(
                local_filename,
            ),
        )

    # If the local file exists and we're not forcing a re-download, no-op
    if local_file_info and not force:
        host.noop('file {0} is already downloaded'.format(filename))
        return

    # Figure out where we're connecting (host or user@host)
    connection_target = hostname
    if ssh_user:
        connection_target = '@'.join((ssh_user, hostname))

    if ssh_keyscan:
        yield keyscan(hostname, state=state, host=host)

    # Download the file with scp
    yield 'scp -P {0} {1}:{2} {3}'.format(
        port, connection_target, filename, local_filename,
    )



from unittest import TestCase

from pyinfra import pseudo_host, pseudo_inventory
from pyinfra.api import Host, Inventory


def _create_host():
    return Host(None, None, None, None)


class TestPseudoHost(TestCase):
    def test_pseudo_host_dir_matches_class(self):
        assert dir(Host) == dir(pseudo_host)

    def test_pseudo_host_eq(self):
        host = _create_host()
        pseudo_host.set(host)
        assert host == pseudo_host

    def test_pseudo_host_repr(self):
        host = _create_host()
        pseudo_host.set(host)
        assert repr(pseudo_host) == 'PseudoModule(Host):Host(None)'

    def test_pseudo_host_str(self):
        host = _create_host()
        pseudo_host.set(host)
        assert str(pseudo_host) == 'None'

    def test_pseudo_host_attr(self):
        host = _create_host()
        pseudo_host.set(host)

        with self.assertRaises(AttributeError):
            host.hello

        setattr(host, 'hello', 'world')
        assert pseudo_host.hello == host.hello

    def test_pseudo_host_class_attr(self):
        host = _create_host()
        pseudo_host.set(host)
        assert pseudo_host.isset() is True

        with self.assertRaises(AttributeError):
            host.hello

        setattr(Host, 'hello', 'class_world')
        setattr(host, 'hello', 'instance_world')

        assert pseudo_host.hello == host.hello

        # Reset and check fallback to class variable
        pseudo_host.reset()
        assert pseudo_host.isset() is False
        assert pseudo_host.hello == 'class_world'


class TestPseudoInventory(TestCase):
    def test_pseudo_inventory_len(self):
        inventory = Inventory(('host', 'anotherhost'))
        pseudo_inventory.set(inventory)
        assert pseudo_inventory.isset() is True

        assert len(pseudo_inventory) == len(inventory)

    def test_pseudo_inventory_iter(self):
        inventory = Inventory(('host', 'anotherhost'))
        pseudo_inventory.set(inventory)
        assert pseudo_inventory.isset() is True

        assert list(iter(pseudo_inventory)) == list(iter(inventory))



from collections import defaultdict

import six

from pyinfra import logger

from .connectors import ALL_CONNECTORS, EXECUTION_CONNECTORS
from .exceptions import NoConnectorError, NoGroupError, NoHostError
from .host import Host
from .util import FallbackDict


def extract_name_data(names):
    for name in names:
        data = {}

        if isinstance(name, tuple):
            data = name[1]
            name = name[0]

        yield name, data


class Inventory(object):
    '''
    Represents a collection of target hosts. Stores and provides access to group data,
    host data and default data for these hosts.

    Args:
        names_data: tuple of ``(names, data)``
        ssh_user: override SSH user
        ssh_port: override SSH port
        ssh_key: override SSH key filename
        ssh_key_password: override password for the SSH key
        ssh_password: override SSH password
        winrm_username: override WINRM username
        winrm_password: override WINRM pasword
        winrm_port: override WINRM port
        winrm_transport: override WINRM transport
        **groups: map of group names -> ``(names, data)``
    '''

    state = None

    def __init__(
        self, names_data,
        ssh_user=None, ssh_port=None, ssh_key=None,
        ssh_key_password=None, ssh_password=None,
        winrm_username=None, winrm_password=None,
        winrm_port=None, winrm_transport=None,
        **groups
    ):
        # Setup basics
        self.groups = defaultdict(list)  # lists of Host objects
        self.host_data = defaultdict(dict)  # dict of name -> data
        self.group_data = defaultdict(dict)  # dict of name -> data

        # In CLI mode these are --user, --key, etc
        override_data = {
            'ssh_user': ssh_user,
            'ssh_key': ssh_key,
            'ssh_key_password': ssh_key_password,
            'ssh_port': ssh_port,
            'ssh_password': ssh_password,
            'winrm_username': winrm_username,
            'winrm_password': winrm_password,
            'winrm_port': winrm_port,
            'winrm_transport': winrm_transport,
        }
        # Strip None values
        override_data = {
            key: value
            for key, value in six.iteritems(override_data)
            if value is not None
        }
        self.override_data = override_data

        names, data = names_data

        # Assign global data
        self.data = data

        # Create the actual host instances and groups
        self.hosts = self.make_hosts_and_groups(names, groups)

    def make_hosts_and_groups(self, names, groups):
        # Map name -> data
        name_to_data = defaultdict(dict)
        # Map name -> group names
        name_to_group_names = defaultdict(list)

        for group_name, (group_names, group_data) in six.iteritems(groups):
            # Assign group data
            self.group_data[group_name] = group_data

            # For any hosts in the group, assign mappings
            for name, data in extract_name_data(group_names):
                name_to_data[name].update(data)
                name_to_group_names[name].append(group_name)

        # Build all/top-level host data - *before* we expand any inventory
        # connectors.
        for name, data in extract_name_data(names):
            name_to_data[name].update(data)

        # Now, use the above to fill self.host_data and populate names_executors
        names_executors = []

        for name, _ in extract_name_data(names):
            host_data = name_to_data[name]

            # Default to executing commands with the ssh connector
            executor = EXECUTION_CONNECTORS['ssh']

            if name[0] == '@':
                connector_name = name[1:]
                arg_string = None

                if '/' in connector_name:
                    connector_name, arg_string = connector_name.split('/', 1)

                if connector_name not in ALL_CONNECTORS:
                    raise NoConnectorError(
                        'Invalid connector: {0}'.format(connector_name),
                    )

                # Execution connector? Simple, just set it for their host
                if connector_name in EXECUTION_CONNECTORS:
                    executor = EXECUTION_CONNECTORS[connector_name]

                names_data = ALL_CONNECTORS[connector_name].make_names_data(arg_string)
                connector_inventory_name = name
            else:
                names_data = [(name, {}, [])]
                connector_inventory_name = None

            for name, data, group_names in names_data:
                # Make a copy of the host data, update with any from
                # the connector.
                sub_host_data = host_data.copy()
                sub_host_data.update(data)

                # Assign the name/data/group_names from the connector
                self.host_data[name] = sub_host_data
                names_executors.append((name, executor))
                name_to_group_names[name].extend(group_names)

                # If we have a connector inventory name, copy any groups attached
                # to the newly generated host name.
                if connector_inventory_name:
                    name_to_group_names[name].extend(
                        name_to_group_names[connector_inventory_name],
                    )

        # Now we can actually make Host instances
        hosts = {}

        for name, executor in names_executors:
            host_groups = name_to_group_names[name]

            # Create the (waterfall data: override, host, group, global)
            host_data = FallbackDict(
                self.get_override_data(),
                self.get_host_data(name),
                self.get_groups_data(host_groups),
                self.get_data(),
                # Pass the method, rather than data, as this comes from the
                # state and can change during deploy(s).
                self.get_deploy_data,
            )

            # Create the Host object
            host = Host(
                name,
                inventory=self,
                groups=name_to_group_names.get(name),
                data=host_data,
                executor=executor,
            )
            hosts[name] = host

            # And push into any groups
            for group_name in host_groups:
                if host not in self.groups[group_name]:
                    self.groups[group_name].append(host)

        return hosts

    def __len__(self):
        '''
        Returns the number of inventory hosts.
        '''

        return len(self.hosts)

    def __iter__(self):
        '''
        Iterates over all inventory hosts.
        '''

        return six.itervalues(self.hosts)

    def len_all_hosts(self):
        logger.warning('`Inventory.len_all_hosts` is deprecated, please use `len(Inventory)`.')
        return len(self)

    def iter_all_hosts(self):
        logger.warning('`Inventory.iter_all_hosts` is deprecated, please use `iter(Inventory)`.')
        return iter(self)

    def iter_active_hosts(self):
        '''
        Iterates over active inventory hosts.
        '''

        return iter(self.state.active_hosts)

    def len_active_hosts(self):
        '''
        Returns the number of active inventory hosts.
        '''

        return len(self.state.active_hosts)

    def iter_activated_hosts(self):
        '''
        Iterates over activated inventory hosts.
        '''

        return iter(self.state.activated_hosts)

    def len_activated_hosts(self):
        '''
        Returns the number of activated inventory hosts.
        '''

        return len(self.state.activated_hosts)

    def get_host(self, name, default=NoHostError):
        '''
        Get a single host by name.
        '''

        if name in self.hosts:
            return self.hosts[name]

        if default is NoHostError:
            raise NoHostError('No such host: {0}'.format(name))

        return default

    def get_group(self, name, default=NoGroupError):
        '''
        Get a list of hosts belonging to a group.
        '''

        if name in self.groups:
            return self.groups[name]

        if default is NoGroupError:
            raise NoGroupError('No such group: {0}'.format(name))

        return default

    def get_data(self):
        '''
        Get the base/all data attached to this inventory.
        '''

        return self.data

    def get_override_data(self):
        '''
        Get override data for this inventory.
        '''

        return self.override_data

    def get_host_data(self, hostname):
        '''
        Get data for a single host in this inventory.
        '''

        return self.host_data.get(hostname, {})

    def get_group_data(self, group):
        '''
        Get data for a single group in this inventory.
        '''

        return self.group_data.get(group, {})

    def get_groups_data(self, groups):
        '''
        Gets aggregated data from a list of groups. Vars are collected in order so, for
        any groups which define the same var twice, the last group's value will hold.
        '''

        data = {}

        for group in groups:
            data.update(self.get_group_data(group))

        return data

    def get_deploy_data(self):
        '''
        Gets any default data attached to the current deploy, if any.
        '''

        if self.state and self.state.deploy_data:
            return self.state.deploy_data

        return {}



import os
import signal
import sys

import click
import gevent

import pyinfra

from .main import cli


# Set CLI mode
pyinfra.is_cli = True

# Don't write out deploy.pyc/config.pyc etc
sys.dont_write_bytecode = True

# Shut it click
click.disable_unicode_literals_warning = True  # noqa

# Force line buffering
sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 1)
sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 1)


def _handle_interrupt(signum, frame):
    click.echo('Exiting upon user request!')
    sys.exit(0)


try:
    # Kill any greenlets on ctrl+c
    gevent.signal_handler(signal.SIGINT, gevent.kill)
except AttributeError:
    # Legacy (gevent <1.2) support
    gevent.signal(signal.SIGINT, gevent.kill)

signal.signal(signal.SIGINT, _handle_interrupt)  # print the message and exit main


if __name__ == 'pyinfra_cli.__main__':
    cli()



import json

from pyinfra.api import FactBase


class LxdContainers(FactBase):
    '''
    Returns a list of running LXD containers
    '''

    command = 'lxc list --format json --fast'
    requires_command = 'lxc'

    default = list

    def process(self, output):
        assert len(output) == 1
        return json.loads(output[0])



import os
import sys
from datetime import datetime
from unittest import TestCase

import pytest
from six.moves import cStringIO as StringIO

from pyinfra.operations import server
from pyinfra_cli.exceptions import CliError
from pyinfra_cli.util import get_operation_and_args, json_encode


class TestCliUtil(TestCase):
    def test_json_encode_function(self):
        assert json_encode(get_operation_and_args) == 'Function: get_operation_and_args'

    def test_json_encode_datetime(self):
        now = datetime.utcnow()
        assert json_encode(now) == now.isoformat()

    def test_json_encode_file(self):
        file = StringIO()
        assert json_encode(file) == 'In memory file: '

    def test_json_encode_set(self):
        assert json_encode({1, 2, 3}) == [1, 2, 3]

    def test_setup_no_module(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('no.op',))
        assert context.exception.message == 'No such module: no'

    def test_setup_no_op(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('server.no',))

        assert context.exception.message == 'No such operation: server.no'

    def test_setup_op_and_args(self):
        commands = ('server.user', 'one', 'two', 'hello=world')

        assert get_operation_and_args(commands) == (
            server.user,
            (['one', 'two'], {'hello': 'world'}),
        )

    def test_setup_op_and_json_args(self):
        commands = ('server.user', '[["one", "two"], {"hello": "world"}]')

        assert get_operation_and_args(commands) == (
            server.user,
            (['one', 'two'], {'hello': 'world'}),
        )


@pytest.fixture(scope='function')
def user_sys_path():
    user_pkg = os.path.dirname(__file__) + '/user'
    sys.path.append(user_pkg)
    yield None
    sys.path.pop()
    to_rm = []
    for k, v in sys.modules.items():
        v = getattr(v, '__file__', '')
        if isinstance(v, str) and v.startswith(user_pkg):
            to_rm.append(k)
    for k in to_rm:
        del sys.modules[k]


# def test_no_user_op():
#     commands = ('test_ops.dummy_op', 'arg1', 'arg2')
#     with pytest.raises(CliError, match='^No such module: test_ops$'):
#         get_operation_and_args(commands)


def test_user_op(user_sys_path):
    commands = ('test_ops.dummy_op', 'arg1', 'arg2')
    res = get_operation_and_args(commands)

    import test_ops
    assert res == (test_ops.dummy_op, (['arg1', 'arg2'], {}))



'''
Manage brew packages on mac/OSX. See https://brew.sh/
'''

from pyinfra.api import operation
from pyinfra.facts.brew import (
    BrewCasks,
    BrewPackages,
    BrewTaps,
    BrewVersion,
    new_cask_cli,
)

from .util.packaging import ensure_packages


@operation(is_idempotent=False)
def update(state=None, host=None):
    '''
    Updates brew repositories.
    '''

    yield 'brew update'

_update = update  # noqa: E305


@operation(is_idempotent=False)
def upgrade(state=None, host=None):
    '''
    Upgrades all brew packages.
    '''

    yield 'brew upgrade'

_upgrade = upgrade  # noqa: E305


@operation
def packages(
    packages=None, present=True, latest=False,
    update=False, upgrade=False,
    state=None, host=None,
):
    '''
    Add/remove/update brew packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``brew update`` before installing packages
    + upgrade: run ``brew upgrade`` before installing packages

    Versions:
        Package versions can be pinned like brew: ``<pkg>@<version>``.

    Examples:

    .. code:: python

        # Update package list and install packages
        brew.packages(
            name='Install Vim and vimpager',
            packages=['vimpager', 'vim'],
            update=True,
        )

        # Install the latest versions of packages (always check)
        brew.packages(
            name='Install latest Vim',
            packages=['vim'],
            latest=True,
        )
    '''

    if update:
        yield _update(state=state, host=host)

    if upgrade:
        yield _upgrade(state=state, host=host)

    yield ensure_packages(
        host, packages, host.get_fact(BrewPackages), present,
        install_command='brew install',
        uninstall_command='brew uninstall',
        upgrade_command='brew upgrade',
        version_join='@',
        latest=latest,
    )


def cask_args(host):
    return ('', ' --cask') if new_cask_cli(host.get_fact(BrewVersion)) else ('cask ', '')


@operation(
    is_idempotent=False,
    pipeline_facts={
        'brew_version': '',
    })
def cask_upgrade(state=None, host=None):
    '''
    Upgrades all brew casks.
    '''

    yield 'brew %supgrade%s' % cask_args(host)


@operation(pipeline_facts={
    'brew_version': '',
})
def casks(
    casks=None, present=True, latest=False, upgrade=False,
    state=None, host=None,
):
    '''
    Add/remove/update brew casks.

    + casks: list of casks to ensure
    + present: whether the casks should be installed
    + latest: whether to upgrade casks without a specified version
    + upgrade: run brew cask upgrade before installing casks

    Versions:
        Cask versions can be pinned like brew: ``<pkg>@<version>``.

    Example:

    .. code:: python

        brew.casks(
            name='Upgrade and install the latest cask',
            casks=['godot'],
            upgrade=True,
            latest=True,
        )

    '''

    if upgrade:
        yield cask_upgrade(state=state, host=host)

    args = cask_args(host)

    yield ensure_packages(
        host, casks, host.get_fact(BrewCasks), present,
        install_command='brew %sinstall%s' % args,
        uninstall_command='brew %suninstall%s' % args,
        upgrade_command='brew %supgrade%s' % args,
        version_join='@',
        latest=latest,
    )


@operation
def tap(src, present=True, state=None, host=None):
    '''
    Add/remove brew taps.

    + src: the name of the tap
    + present: whether this tap should be present or not

    Examples:

    .. code:: python

        brew.tap(
            name='Add a brew tap',
            src='includeos/includeos',
        )

        # multiple taps
        taps = ['includeos/includeos', 'ktr0731/evans']
        for tap in taps:
            brew.tap(
                name={f'Add brew tap {tap}'},
                src=tap,
            )

    '''

    taps = host.get_fact(BrewTaps)
    is_tapped = src in taps

    if present:
        if is_tapped:
            host.noop('tap {0} already exists'.format(src))
        else:
            yield 'brew tap {0}'.format(src)
            taps.append(src)

    elif not present:
        if is_tapped:
            yield 'brew untap {0}'.format(src)
            taps.remove(src)
        else:
            host.noop('tap {0} does not exist'.format(src))



'''
Manage pacman packages. (Arch Linux package manager)
'''

from pyinfra.api import operation
from pyinfra.facts.pacman import PacmanPackages, PacmanUnpackGroup

from .util.packaging import ensure_packages


@operation
def upgrade(state=None, host=None):
    '''
    Upgrades all pacman packages.
    '''

    yield 'pacman --noconfirm -Su'

_upgrade = upgrade  # noqa: E305


@operation
def update(state=None, host=None):
    '''
    Updates pacman repositories.
    '''

    yield 'pacman -Sy'

_update = update  # noqa: E305


@operation
def packages(
    packages=None, present=True,
    update=False, upgrade=False,
    state=None, host=None,
):
    '''
    Add/remove pacman packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + update: run ``pacman -Sy`` before installing packages
    + upgrade: run ``pacman -Su`` before installing packages

    Versions:
        Package versions can be pinned like pacman: ``<pkg>=<version>``.

    Example:

    .. code:: python

        pacman.packages(
            name='Install Vim and a plugin',
            packages=['vim-fugitive', 'vim'],
            update=True,
        )
    '''

    if update:
        yield _update(state=state, host=host)

    if upgrade:
        yield _upgrade(state=state, host=host)

    yield ensure_packages(
        host, packages, host.get_fact(PacmanPackages), present,
        install_command='pacman --noconfirm -S',
        uninstall_command='pacman --noconfirm -R',
        lower=False,
        expand_package_fact=lambda package: host.get_fact(
            PacmanUnpackGroup,
            name=package,
        ),
    )



'''
Manage systemd services.
'''

from __future__ import unicode_literals

from pyinfra.api import operation
from pyinfra.facts.systemd import SystemdEnabled, SystemdStatus

from .util.service import handle_service_control


@operation
def daemon_reload(user_mode=False, state=None, host=None):
    '''
    Reload the systemd daemon to read unit file changes.

    + user_mode: whether to use per-user systemd (systemctl --user) or not
    '''

    systemctl_cmd = 'systemctl --user' if user_mode else 'systemctl'
    yield '{0} daemon-reload'.format(systemctl_cmd)

_daemon_reload = daemon_reload  # noqa: E305


@operation
def service(
    service,
    running=True, restarted=False, reloaded=False,
    command=None, enabled=None, daemon_reload=False,
    user_mode=False, state=None, host=None,
):
    '''
    Manage the state of systemd managed units.

    + service: name of the systemd unit to manage
    + running: whether the unit should be running
    + restarted: whether the unit should be restarted
    + reloaded: whether the unit should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<name> <command>``
    + enabled: whether this unit should be enabled/disabled on boot
    + daemon_reload: reload the systemd daemon to read updated unit files
    + user_mode: whether to use per-user systemd (systemctl --user) or not

    Example:

    .. code:: python

        systemd.service(
            name='Restart and enable the dnsmasq service',
            service='dnsmasq.service',
            running=True,
            restarted=True,
            enabled=True,
        )

        systemd.service(
            name='Enable logrotate timer',
            service='logrotate.timer',
            running=True,
            enabled=True,
        )

    '''

    systemctl_cmd = 'systemctl --user' if user_mode else 'systemctl'

    if '.' not in service:
        service = '{0}.service'.format(service)

    if daemon_reload:
        yield _daemon_reload(user_mode=user_mode, state=state, host=host)

    yield handle_service_control(
        host,
        service, SystemdStatus,
        ' '.join([systemctl_cmd, '{1}', '{0}']),
        running, restarted, reloaded, command,
    )

    if isinstance(enabled, bool):
        systemd_enabled = host.get_fact(SystemdEnabled)
        is_enabled = systemd_enabled.get(service, False)

        # Isn't enabled and want enabled?
        if not is_enabled and enabled is True:

            yield ' '.join([systemctl_cmd, 'enable', '{0}']).format(service)
            systemd_enabled[service] = True

        # Is enabled and want disabled?
        elif is_enabled and enabled is False:
            yield ' '.join([systemctl_cmd, 'disable', '{0}']).format(service)
            systemd_enabled[service] = False



'''
The pyinfra facts API. Facts enable pyinfra to collect remote server state which
is used to "diff" with the desired state, producing the final commands required
for a deploy.
'''

from __future__ import division, unicode_literals

import re

from inspect import getcallargs, isclass
from socket import (
    error as socket_error,
    timeout as timeout_error,
)

import click
import gevent
import six

from gevent.lock import BoundedSemaphore
from paramiko import SSHException

from pyinfra import logger
from pyinfra.api import StringCommand
from pyinfra.api.connectors.util import split_combined_output
from pyinfra.api.util import (
    get_arg_value,
    get_kwargs_str,
    log_error_or_warning,
    log_host_command_error,
    make_hash,
    print_host_combined_output,
    underscore,
)
from pyinfra.progress import progress_spinner

from .operation_kwargs import get_executor_kwarg_keys


# Index of snake_case facts -> CamelCase classes
FACTS = {}
FACT_LOCK = BoundedSemaphore()

SUDO_REGEX = r'^sudo: unknown user:'
SU_REGEXES = (
    r'^su: user .+ does not exist',
    r'^su: unknown login',
)


def is_fact(name):
    return name in FACTS


def get_fact_class(name):
    return FACTS[name]


def get_fact_names():
    '''
    Returns a list of available facts in camel_case format.
    '''

    return list(six.iterkeys(FACTS))


class FactMeta(type):
    '''
    Metaclass to dynamically build the facts index.
    '''

    def __init__(cls, name, bases, attrs):
        if attrs.get('abstract'):
            return

        fact_name = underscore(name)
        cls.name = fact_name

        # Get the an instance of the fact, attach to facts
        FACTS[fact_name] = cls


@six.add_metaclass(FactMeta)
class FactBase(object):
    abstract = True

    shell_executable = None

    requires_command = None

    @staticmethod
    def default():
        '''
        Set the default attribute to be a type (eg list/dict).
        '''

    @staticmethod
    def process(output):
        return '\n'.join(output)

    def process_pipeline(self, args, output):
        return {
            arg: self.process([output[i]])
            for i, arg in enumerate(args)
        }


@six.add_metaclass(FactMeta)
class ShortFactBase(object):
    fact = None


def get_short_facts(state, short_fact, **kwargs):
    facts = get_facts(state, short_fact.fact.name, **kwargs)

    return {
        host: short_fact.process_data(data)
        for host, data in six.iteritems(facts)
    }


def _make_command(command_attribute, host_args):
    if callable(command_attribute):
        host_args.pop('self', None)
        return command_attribute(**host_args)
    return command_attribute


def get_facts(
    state,
    name_or_cls,
    args=None,
    kwargs=None,
    ensure_hosts=None,
    apply_failed_hosts=True,
):
    '''
    Get a single fact for all hosts in the state.
    '''

    # TODO: tidy up the whole executor argument handling here!
    global_kwarg_overrides = {}
    if kwargs:
        global_kwarg_overrides.update({
            key: kwargs.pop(key)
            for key in get_executor_kwarg_keys()
            if key in kwargs
        })

    if isclass(name_or_cls) and issubclass(name_or_cls, (FactBase, ShortFactBase)):
        fact = name_or_cls()
        name = fact.name
    else:
        fact = get_fact_class(name_or_cls)()
        name = name_or_cls

    if isinstance(fact, ShortFactBase):
        return get_short_facts(state, fact, args=args, ensure_hosts=ensure_hosts)

    args = args or ()
    kwargs = kwargs or {}
    if args or kwargs:
        # Merges args & kwargs into a single kwargs dictionary
        kwargs = getcallargs(fact.command, *args, **kwargs)

    logger.debug('Getting fact: {0} ({1}) (ensure_hosts: {2})'.format(
        name, get_kwargs_str(kwargs), ensure_hosts,
    ))

    # Apply args or defaults
    sudo = state.config.SUDO
    sudo_user = state.config.SUDO_USER
    su_user = state.config.SU_USER
    ignore_errors = state.config.IGNORE_ERRORS
    shell_executable = state.config.SHELL
    use_sudo_password = state.config.USE_SUDO_PASSWORD
    env = state.config.ENV

    # Facts can override the shell (winrm powershell vs cmd support)
    if fact.shell_executable:
        shell_executable = fact.shell_executable

    # Timeout for operations !== timeout for connect (config.CONNECT_TIMEOUT)
    timeout = None

    # If inside an operation, fetch global arguments
    current_global_kwargs = state.current_op_global_kwargs or {}
    # Allow `Host.get_fact` calls to explicitly override these
    current_global_kwargs.update(global_kwarg_overrides)
    if current_global_kwargs:
        sudo = current_global_kwargs.get('sudo', sudo)
        sudo_user = current_global_kwargs.get('sudo_user', sudo_user)
        use_sudo_password = current_global_kwargs.get('use_sudo_password', use_sudo_password)
        su_user = current_global_kwargs.get('su_user', su_user)
        ignore_errors = current_global_kwargs.get('ignore_errors', ignore_errors)
        timeout = current_global_kwargs.get('timeout', timeout)
        env = current_global_kwargs.get('env', env)

    # Make a hash which keeps facts unique - but usable cross-deploy/threads.
    # Locks are used to maintain order.
    fact_hash = make_hash((name, kwargs, sudo, sudo_user, su_user, ignore_errors, env))

    # Already got this fact? Unlock and return them
    current_facts = state.facts.get(fact_hash, {})
    if current_facts:
        if not ensure_hosts or all(
            host in current_facts for host in ensure_hosts
        ):
            return current_facts

    with FACT_LOCK:
        # Add any hosts we must have, whether considered in the inventory or not
        # (these hosts might be outside the --limit or current op limit_hosts).
        hosts = set(state.inventory.iter_active_hosts())
        if ensure_hosts:
            hosts.update(ensure_hosts)

        # Execute the command for each state inventory in a greenlet
        greenlet_to_host = {}

        for host in hosts:
            if host in current_facts:
                continue

            # Generate actual arguments by passing strings as jinja2 templates
            host_kwargs = {
                key: get_arg_value(state, host, arg)
                for key, arg in kwargs.items()
            }

            command = _make_command(fact.command, host_kwargs)
            requires_command = _make_command(fact.requires_command, host_kwargs)
            if requires_command:
                command = StringCommand(
                    # Command doesn't exist, return 0 *or* run & return fact command
                    '!', 'command', '-v', requires_command, '>/dev/null', '||', command,
                )

            greenlet = state.fact_pool.spawn(
                host.run_shell_command,
                command,
                sudo=sudo,
                sudo_user=sudo_user,
                use_sudo_password=use_sudo_password,
                su_user=su_user,
                timeout=timeout,
                env=env,
                shell_executable=shell_executable,
                print_output=state.print_fact_output,
                print_input=state.print_fact_input,
                return_combined_output=True,
            )
            greenlet_to_host[greenlet] = host

        # Wait for all the commands to execute
        progress_prefix = 'fact: {0} ({1})'.format(name, get_kwargs_str(kwargs))

        with progress_spinner(
            greenlet_to_host.values(),
            prefix_message=progress_prefix,
        ) as progress:
            for greenlet in gevent.iwait(greenlet_to_host.keys()):
                host = greenlet_to_host[greenlet]
                progress(host)

        hostname_facts = {}
        failed_hosts = set()

        # Collect the facts and any failures
        for greenlet, host in six.iteritems(greenlet_to_host):
            status = False
            stdout = []

            try:
                status, combined_output_lines = greenlet.get()
            except (timeout_error, socket_error, SSHException) as e:
                failed_hosts.add(host)
                log_host_command_error(
                    host, e,
                    timeout=timeout,
                )

            stdout, stderr = split_combined_output(combined_output_lines)

            data = fact.default()

            if status:
                if stdout:
                    data = fact.process(stdout)

            elif stderr:
                first_line = stderr[0]
                if (
                    sudo_user
                    and re.match(SUDO_REGEX, first_line)
                ):
                    status = True
                if (
                    su_user
                    and any(re.match(regex, first_line) for regex in SU_REGEXES)
                ):
                    status = True

            if not status:
                failed_hosts.add(host)

                if not state.print_fact_output:
                    print_host_combined_output(host, combined_output_lines)

                log_error_or_warning(host, ignore_errors, description=(
                    'could not load fact: {0} {1}'
                ).format(name, get_kwargs_str(kwargs)))

            hostname_facts[host] = data

        log = 'Loaded fact {0} ({1})'.format(click.style(name, bold=True), get_kwargs_str(kwargs))
        if state.print_fact_info:
            logger.info(log)
        else:
            logger.debug(log)

        # Check we've not failed
        if not ignore_errors and apply_failed_hosts:
            state.fail_hosts(failed_hosts)

        # Assign the facts
        state.facts.setdefault(fact_hash, {}).update(hostname_facts)

    return state.facts[fact_hash]


def get_host_fact(state, host, name, args=None, kwargs=None):
    fact_data = get_facts(state, name, args=args, kwargs=kwargs, ensure_hosts=(host,))
    return fact_data.get(host)


def create_host_fact(state, host, name, data, args=None, kwargs=None):
    fact_data = get_facts(state, name, args=args, kwargs=kwargs, ensure_hosts=(host,))
    fact_data[host] = data


def delete_host_fact(state, host, name, args=None, kwargs=None):
    fact_data = get_facts(state, name, args=args, kwargs=kwargs, ensure_hosts=(host,))
    fact_data.pop(host, None)
